#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Module containing types and utilities for building generator objects.
Generators abstract the code generation technology used to transform
pydsdl AST into source code.
"""

import abc
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping, Optional, Type, Union

from ._namespace import Generatable, Namespace
from ._utilities import ResourceType, YesNoDefault
from .lang import LanguageContext, LanguageContextBuilder
from .lang._language import Language


@dataclass(frozen=True)
class GenerationResult:
    """
    A simple data class to hold the results of a generation operation.
    """

    lctx: LanguageContext
    """The language context used to generate outputs."""

    generator_targets: Dict[Path, Generatable]
    """The set of explicit files targeted for generation and their dependant files."""

    generated_files: List[Path]
    """
    The set of files that were (or would be, for dry runs) generated for data types.
    """

    support_files: List[Path]
    """
    The set of files that were (or would be, for dry runs) generated to support data types.
    """

    template_files: List[Path]
    """
    The set of template files used to generate the `generated_files`.
    """

    def __add__(self, other: Any) -> "GenerationResult":
        """
        If there exists an isomorphism between this object and other, return a union of the two as a new result.

        .. invisible-code-block: python

            from nunavut._generators import GenerationResult, basic_language_context_builder_from_args
            from pathlib import Path
            from pytest import raises

            c_lang = basic_language_context_builder_from_args(target_language="c").create()
            js_lang = basic_language_context_builder_from_args(
                target_language="js",
                include_experimental_languages=True).create()

            with raises(RuntimeError):
                GenerationResult(c_lang, {}, [], [], []) + 1

            with raises(RuntimeError):
                GenerationResult(c_lang, {}, [], [], []) + GenerationResult(js_lang, {}, [], [], [])

        .. code-block:: python

            ursine = GenerationResult(c_lang,
                                      {Path("bears/grizzly.c"): Path("bears/grizzly.dsdl")},
                                      [Path("bears/grizzly.c")],
                                      [Path("include/support.h")],
                                      [Path("templates/code.j2")])
            bovine = GenerationResult(c_lang,
                                      {Path("cows/jersey.c"): Path("cows/jersey.dsdl")},
                                      [Path("cows/jersey.c")],
                                      [Path("include/support.h")],
                                      [Path("templates/code.j2")])

            mammals = ursine + bovine

            assert(mammals.generator_targets[Path("bears/grizzly.c")] == Path("bears/grizzly.dsdl"))
            assert(mammals.generator_targets[Path("cows/jersey.c")] == Path("cows/jersey.dsdl"))
            assert(len(mammals.support_files) == 1)
            assert(len(mammals.template_files) == 1)

        """
        if not isinstance(other, self.__class__):
            raise RuntimeError(f"Cannot add {type(other)} type to a GenerationResult.")
        if other.lctx != self.lctx:
            raise RuntimeError(
                f"Result with language {str(other.lctx)} is not isomorphic with this result for "
                f"language {str(self.lctx)}."
            )

        return GenerationResult(
            self.lctx,
            {**self.generator_targets, **other.generator_targets},
            self.generated_files + other.generated_files,
            [*{*(self.support_files + other.support_files)}],
            [*{*(self.template_files + other.template_files)}],
        )


class AbstractGenerator(metaclass=abc.ABCMeta):
    """
    Abstract base class for classes that generate source file output
    from a given pydsdl parser result.

    :param nunavut.Namespace namespace:  The top-level namespace to
        generates types at and from.
    :param int resource_types: A bitmask of resources to generate.
        This can be a combination of ResourceType values.
    :param YesNoDefault generate_namespace_types:  Set to YES
        to force generation files for namespaces and NO to suppress.
        DEFAULT will generate namespace files based on the language
        preference.
    :param Iterable[Path] index_file: A list of paths to files that
        should be generated, relative to the output directory, which
        are given access to the full namespace context rather than
        per-type context.
    :param Any kwargs: Additional arguments to pass into generators.
    """

    def __init__(
        self,
        namespace: Namespace,
        resource_types: int,
        generate_namespace_types: YesNoDefault = YesNoDefault.DEFAULT,
        index_file: Optional[Iterable[Path]] = None,
        **kwargs: Any,
    ):  # pylint: disable=unused-argument
        self._namespace = namespace
        self._resource_types = resource_types
        self._generate_namespace_types = self.generate_namespace_types_from_trinary(
            self._namespace.get_language_context().get_target_language(), generate_namespace_types
        )
        if index_file is not None:
            self._index_files = [Path(p) for p in index_file]
        else:
            self._index_files = []

    @classmethod
    def generate_namespace_types_from_trinary(
        cls, target_language: Language, generate_namespace_types: YesNoDefault
    ) -> bool:
        """
        Given the target language and a trinary value, returns a binary result for "should namespace types be generated"
        as a parameter.
        """
        if generate_namespace_types == YesNoDefault.YES:
            return True
        elif generate_namespace_types == YesNoDefault.NO:
            return False
        else:
            if target_language.has_standard_namespace_files:
                return True
            else:
                return False

    @property
    def namespace(self) -> Namespace:
        """
        The root :class:`nunavut.Namespace` for this generator.
        """
        return self._namespace

    @property
    def resource_types(self) -> int:
        """
        The bitmask of resources to generate. This can be a combination of ResourceType values.
        """
        return self._resource_types

    @property
    def generate_namespace_types(self) -> bool:
        """
        If true then the generator is set to emit files for :class:`nunavut.Namespace`
        in addition to the pydsdl datatypes. If false then only files for pydsdl datatypes
        will be generated.
        """
        return self._generate_namespace_types

    @property
    def index_files(self) -> List[Path]:
        """
        A list of paths to files that should be generated, relative to the output directory, which are given access
        to the full namespace context rather than per-type context.
        """
        return self._index_files

    @abc.abstractmethod
    def get_templates(self) -> Iterable[Path]:
        """
        Enumerate all templates found in the templates path.
        :return: A list of paths to all templates found by this Generator object.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def generate_all(
        self,
        is_dryrun: bool = False,
        allow_overwrite: bool = True,
    ) -> Iterable[Path]:
        """
        Generates all output for a given :class:`nunavut.Namespace` and using
        the templates found by this object.

        :param bool is_dryrun: If True then no output files will actually be
                               written but all other operations will be performed.
        :param bool allow_overwrite: If True then the generator will attempt to overwrite any existing files
                                it encounters. If False then the generator will raise an error if the
                                output file exists and the generation is not a dry-run.
        :return: Iterator over the files generated.
        :raises: PermissionError if :attr:`allow_overwrite` is False and the file exists.
        """
        raise NotImplementedError()


# --[GENERATION HELPERS]-------------------------------------------------------


def basic_language_context_builder_from_args(target_language: str, **kwargs: Any) -> LanguageContextBuilder:
    """
    Uses interpreted arguments to create a new LanguageContextBuilder object with support for the most common arguments.
    This method will always support arguments used in the CLI. For more advanced use cases, create a
    LanguageContextBuilder object directly or add additional arguments to the builder after creation.

    .. note:: This is the same method used by :func:`generate_all` to create a LanguageContextBuilder object.

    .. invisible-code-block: python

        from nunavut import basic_language_context_builder_from_args

    .. code-block:: python

        # target_language is the only required argument
        builder = basic_language_context_builder_from_args(target_language="c")

        assert builder.create().get_target_language().name == "c"

    :param str target_language: The name of the language to generate source for.
    :param kwargs: A dictionary of arguments to pass to the :class:`LanguageContextBuilder`.
        The following arguments are supported:

            * **configuration**: A list of additional configuration files to load.
            * **option**: Key value arguments to override individual language option values.
            * **include_experimental_languages**: If true then experimental languages will also be available.
            * **language_options**: Opaque arguments passed through to the language objects. The supported arguments and
                valid values are different depending on the language specified by the :code:`language_key` parameter.
            * **output_extension**: The file extension to use for generated files.
            * **namespace_output_stem**: The stem to use for generated namespace files.
            * **additional_config_files**: A list of :class:`Path` s to additional configuration files to load.

    :return: A new :class:`LanguageContextBuilder` object based on the command line arguments.
    """

    additional_config_files = kwargs.get("configuration")
    if additional_config_files is None:
        additional_config_files = []
    if isinstance(additional_config_files, Path):
        additional_config_files = [additional_config_files]

    builder = (
        LanguageContextBuilder(include_experimental_languages=kwargs.get("include_experimental_languages", False))
        .set_target_language(target_language)
        .set_target_language_configuration_override(
            Language.WKCV_LANGUAGE_OPTIONS, kwargs.get("language_options", None)
        )
        .set_target_language_extension(kwargs.get("output_extension", None))
        .set_target_language_configuration_override(
            Language.WKCV_NAMESPACE_FILE_STEM, kwargs.get("namespace_output_stem", None)
        )
        .add_config_files(*additional_config_files)
    )

    options = kwargs.get("option")
    if options is not None:
        for option_key, option_value in options:
            builder.add_target_language_option_override(option_key, option_value)

    return builder


# --[GENERATION FUNCTIONS]-----------------------------------------------------


# pylint: disable=too-many-arguments, too-many-locals, too-many-statements
def generate_all(
    target_language: str,
    target_files: Iterable[Union[str, Path]],
    root_namespace_directories_or_names: Iterable[Union[str, Path]],
    outdir: Path,
    language_options: Optional[Mapping[str, Any]] = None,
    include_experimental_languages: bool = False,
    resource_types: int = ResourceType.ANY.value,
    embed_auditing_info: bool = False,
    dry_run: bool = False,
    jobs: int = 0,
    no_overwrite: bool = False,
    allow_unregulated_fixed_port_id: bool = False,
    omit_dependencies: bool = False,
    code_generator_type: Optional[Type[AbstractGenerator]] = None,
    support_generator_type: Optional[Type[AbstractGenerator]] = None,
    **kwargs: Any,
) -> GenerationResult:
    """
    Helper method that uses default settings and built-in templates to generate types for a given
    language. This method is the most direct way to generate code using Nunavut.

    Generation takes place four steps with a function provided for each step:


    .. code-block:: none

        ┌───────────────────────────────────────────────────┐ 1. generate_all
        │           Language context construction           │
        │                                                   │
        │  ┌─────────────────────────────────────────────┐  │ 2. generate_all_for_language
        │  │ Parsing (pydsdl) and constructing Namespace │  │
        │  │             trees from results              │  │
        │  │                                             │  │
        │  │  ┌───────────────────────────────────────┐  │  │ 3. generate_all_from_namespace
        │  │  │        Generator construction         │  │  │
        │  │  │                                       │  │  │
        │  │  │  ┌─────────────────────────────────┐  │  │  │ 4. generate_all_from_namespace_with_generators
        │  │  │  │        code generation          │  │  │  │
        │  │  │  │                                 │  │  │  │
        │  │  │  │                                 │  │  │  │
        │  │  │  └─────────────────────────────────┘  │  │  │
        │  │  └───────────────────────────────────────┘  │  │
        │  └─────────────────────────────────────────────┘  │
        └───────────────────────────────────────────────────┘

    At each stage the number of options are reduced as objects are constructed based on their values.

    * **generate_all**: Requires no Nunavut objects and can be driven from a command-line.
    * **generate_all_for_language**: This is the best entry point for programmatic use of Nunavut. Simply construct a
        LanguageContext object using a LanguageContextBuilder and pass it to this function.
    * **generate_all_from_namespace**: This function can be used if you have already constructed a Namespace object.
    * **generate_all_from_namespace_with_generators**: This function is used if you have already constructed the
        generators you want to use.

    :param str target_language: The name of the language to generate source for.
                See the :doc:`../../docs/templates` for details on available language support.
    :param target_files: A list of paths to dsdl files. This method will generate code for these files and their
                dependant types.
    :param root_namespace_directories_or_names: This can be a set of names of root namespaces or relative paths to
        root namespaces. All ``dsdl_files`` provided must be under one of these roots. For example, given:

        .. code-block:: python

            target_dsdl_files = [
                            Path("workspace/project/types/animals/felines/Tabby.1.0.dsdl"),
                            Path("workspace/project/types/animals/canines/Boxer.1.0.dsdl"),
                            Path("workspace/project/types/plants/trees/DouglasFir.1.0.dsdl")
                        ]


        then this argument must be one of:

        .. code-block:: python

            root_namespace_directories_or_names = ["animals", "plants"]

            # or

            root_namespace_directories_or_names = [
                            Path("workspace/project/types/animals"),
                            Path("workspace/project/types/plants")
                        ]


    :param Path outdir:
        The path to generate code at and under.
    :param Optional[Mapping[str, Any]] language_options: Opaque arguments passed through to the language objects. The
        supported arguments and valid values are different depending on the language specified by the `language_key`
        parameter.
    :param int resource_types:
        Bitmask of resources to generate. This can be a combination of ResourceType values. For example, to generate
        only serialization support code, set this to ResourceType.SERIALIZATION_SUPPORT.value. To generate all resources
        set this to ResourceType.ANY.value. To only generate resource files (i.e. to omit source code generation) set
        the ResourceType.ONLY.value bit and the resource type bytes you do want to generate.
    :param embed_auditing_info:
        If True then additional information about the inputs and environment used to generate source will be embedded in
        the generated files at the cost of build reproducibility.
    :param bool include_experimental_languages:
        If true then experimental languages will also be available.
    :param bool dry_run:
        If True then no files will be generated/written but all logic will be exercised with commensurate logging and
        errors.
    :param int jobs:
        The number of parallel jobs to use when generating code. If 1 then no parallelism is used. If 0 then the
        number of jobs is determined by the number of CPUs available.

        .. note:: By default, any multiprocessing jobs used will not have a timeout set. To set a timeout for any jobs
            set the environment variable ``NUNAVUT_JOB_TIMEOUT_SECONDS`` to the desired timeout in fractional seconds.
            this is normally not useful as a correct timeout value is highly dependent on the system and the number of
            types being generated.

    :param bool no_overwrite:
        If True then generated files will not be allowed to overwrite existing files under the `outdir` path causing
        errors.
    :param bool allow_unregulated_fixed_port_id:
        If True then errors will become warning when using fixed port identifiers for unregulated datatypes.
    :param bool omit_dependencies:
        If True then only the types explicitly provided in `target_files` will be generated. If False then all
        dependant types will also be generated.
    :param code_generator_type: The type of code generator to use. If None then the default code generator is used.
    :param support_generator_type: The type of support generator to use. If None then the default support generator is
        used.
    :param kwargs: Additional arguments passed into the language context builder and generator constructors. See the
        documentation for :func:`basic_language_context_builder_from_args` and for specific generator types for details
        on supported arguments.

    :returns GenerationResult: A dataclass containing explicit inputs, discovered inputs, and determined outputs.
    :raises pydsdl.FrontendError: Exceptions thrown from the pydsdl frontend. For example, parsing malformed DSDL will
        raise this exception.
    """

    language_context = basic_language_context_builder_from_args(
        target_language,
        language_options=language_options,
        include_experimental_languages=include_experimental_languages,
        **kwargs,
    ).create()

    return generate_all_for_language(
        language_context,
        target_files,
        root_namespace_directories_or_names,
        outdir,
        resource_types,
        embed_auditing_info,
        dry_run,
        jobs,
        no_overwrite,
        allow_unregulated_fixed_port_id,
        omit_dependencies,
        code_generator_type,
        support_generator_type,
        **kwargs,
    )


def generate_all_for_language(
    language_context: LanguageContext,
    target_files: Iterable[Union[str, Path]],
    root_namespace_directories_or_names: Iterable[Union[str, Path]],
    outdir: Path,
    resource_types: int = ResourceType.ANY.value,
    embed_auditing_info: bool = False,
    dry_run: bool = False,
    jobs: int = 0,
    no_overwrite: bool = False,
    allow_unregulated_fixed_port_id: bool = False,
    omit_dependencies: bool = False,
    code_generator_type: Optional[Type[AbstractGenerator]] = None,
    support_generator_type: Optional[Type[AbstractGenerator]] = None,
    **generator_args: Any,
) -> GenerationResult:
    """
    Generate code for a set of target_files using a configured language context. This is the second step in the
    generation process and is the best entry point for programmatic use of Nunavut. See the documentation for
    :func:`generate_all` for more details.

    :param language_context: The language context to generate code for.
    :param target_files: A list of paths to dsdl files. This method will generate code for these files and their
        dependant types.
    :param root_namespace_directories_or_names: This can be a set of names of root namespaces or relative paths to
        root namespaces. All ``dsdl_files`` provided must be under one of these roots. See the documentation for
        :func:`generate_all` for more details.
    :param outdir: The path to generate code at and under.
    :param resource_types: A bitmask of resources to generate. This can be a combination of ResourceType values.
    :param embed_auditing_info: If True then additional information about the inputs and environment used to generate
        source will be embedded in the generated files at the cost of build reproducibility.
    :param dry_run: If True then no files will be generated/written but all logic will be exercised with commensurate
        logging and errors.
    :param int jobs:
        The number of parallel jobs to use when generating code. If 1 then no parallelism is used. If 0 then the
        number of jobs is determined by the number of CPUs available.

        .. note:: By default, any multiprocessing jobs used will not have a timeout set. To set a timeout for any jobs
            set the environment variable ``NUNAVUT_JOB_TIMEOUT_SECONDS`` to the desired timeout in fractional seconds.
            this is normally not useful as a correct timeout value is highly dependent on the system and the number of
            types being generated.

    :param no_overwrite: If True then generated files will not be allowed to overwrite existing files under the `outdir`
        path causing errors.
    :param allow_unregulated_fixed_port_id: If True then errors will become warning when using fixed port identifiers
        for unregulated datatypes.
    :param omit_dependencies: If True then only the types explicitly provided in `target_files` will be generated. If
        False then all dependant types will also be generated.
    :param code_generator_type: The type of code generator to use. If None then the default code generator is used.
    :param support_generator_type: The type of support generator to use. If None then the default support generator is
        used.
    :param generator_args: Additional arguments to pass into the generator constructors. See the documentation for
        specific generator types for details on supported arguments.
    :return: A dataclass containing explicit inputs, discovered inputs, and determined outputs.
    :raises pydsdl.FrontendError: Exceptions thrown from the pydsdl frontend. For example, parsing malformed DSDL will
        raise this exception.
    """
    index = Namespace.read_files(
        outdir,
        language_context,
        target_files,
        root_namespace_directories_or_names,
        jobs,
        float(os.environ.get("NUNAVUT_JOB_TIMEOUT_SECONDS", 0)),
        allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id,
        omit_dependencies=omit_dependencies,
    )

    return generate_all_from_namespace(
        index,
        resource_types,
        embed_auditing_info,
        dry_run,
        no_overwrite,
        code_generator_type,
        support_generator_type,
        **generator_args,
    )


def generate_all_from_namespace(
    index: Namespace,
    resource_types: int = ResourceType.ANY.value,
    embed_auditing_info: bool = False,
    dry_run: bool = False,
    no_overwrite: bool = False,
    code_generator_type: Optional[Type[AbstractGenerator]] = None,
    support_generator_type: Optional[Type[AbstractGenerator]] = None,
    **generator_args: Any,
) -> GenerationResult:
    """
    Given a populated namespace, generate code for the types found within it. This is the third step in the generation
    process and is used when you have already constructed a Namespace object. See the documentation for
    :func:`generate_all` for more details.

    :param index: The namespace tree to generate code for.
    :param resource_types: A bitmask of resources to generate. This can be a combination of ResourceType values.
    :param embed_auditing_info: If True then additional information about the inputs and environment used to generate
        source will be embedded in the generated files at the cost of build reproducibility.
    :param dry_run: If True then no files will be generated/written but all logic will be exercised with commensurate
        logging and errors.
    :param no_overwrite: If True then generated files will not be allowed to overwrite existing files under the `outdir`
        path causing errors.
    :param code_generator_type: The type of code generator to use. If None then the default code generator is used.
    :param support_generator_type: The type of support generator to use. If None then the default support generator is
        used.
    :param generator_args: Additional arguments to pass into the generator constructors. See the documentation for
        specific generator types for details on supported arguments.
    :return: A dataclass containing explicit inputs, discovered inputs, and determined outputs.
    :raises pydsdl.FrontendError: Exceptions thrown from the pydsdl frontend. For example, parsing malformed DSDL will
        raise this exception.
    """

    support_generator_args = generator_args.copy()
    support_generator_args["templates_dir"] = generator_args.get("support_templates_dir", [])

    # if ResourceType.INDEX.value() | resource_types != 0:
    #     For implementing issue #334 or other features requiring an index template, add an index_generator_type
    #     here to the generators we create and run

    if code_generator_type is None:
        # load default code generator
        from .jinja import DSDLCodeGenerator  # pylint: disable=import-outside-toplevel

        code_generator_type = DSDLCodeGenerator
    if support_generator_type is None:
        # load default support generator
        from .jinja import SupportGenerator  # pylint: disable=import-outside-toplevel

        support_generator_type = SupportGenerator

    code_generator = code_generator_type(
        index, resource_types, embed_auditing_info=embed_auditing_info, **generator_args
    )
    support_generator = support_generator_type(
        index, resource_types, embed_auditing_info=embed_auditing_info, **support_generator_args
    )

    return generate_all_from_namespace_with_generators(index, code_generator, support_generator, dry_run, no_overwrite)


def generate_all_from_namespace_with_generators(
    index: Namespace,
    code_generator: AbstractGenerator,
    support_generator: AbstractGenerator,
    dry_run: bool = False,
    no_overwrite: bool = False,
) -> GenerationResult:
    """
    Given a populated namespace, generate code for the types found within it using the provided generators. This is the
    fourth and final step in the generation process and is used when you have already constructed the generators you
    want to use. See the documentation for :func:`generate_all` for more details.

    :param index: The namespace tree to generate code for.
    :param code_generator: The code generator to use.
    :param support_generator: The support generator to use.
    :param dry_run: If True then no files will be generated/written but all logic will be exercised with commensurate
        logging and errors.
    :param no_overwrite: If True then generated files will not be allowed to overwrite existing files under the `outdir`
        path causing errors.
    :return: A dataclass containing explicit inputs, discovered inputs, and determined outputs.
    :raises pydsdl.FrontendError: Exceptions thrown from the pydsdl frontend. For example, parsing malformed DSDL will
        raise this exception.
    """

    template_files = list(set(support_generator.get_templates()).union(set(code_generator.get_templates())))
    generated_files = list(code_generator.generate_all(dry_run, not no_overwrite))

    if support_generator.resource_types & ResourceType.ONLY.value != 0 or len(generated_files) > 0:
        # if we are specifically generating support files or if we are generating them as-needed and we need them.
        support_files = list(support_generator.generate_all(dry_run, not no_overwrite))
    else:
        # else we don't need to generate support files because there's nothing to support
        support_files = []

    generator_targets = {}
    for file in generated_files:
        if isinstance(file, Generatable):
            generator_targets[Path(file)] = file

    return GenerationResult(
        index.get_language_context(),
        generator_targets,
        generated_files,
        support_files,
        template_files,
    )



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.

"""
Contains the Jinja templates to generate Py headers.
"""

__version__ = "1.0.0"
"""Version of the py template set."""



#
# Copyright (C) 2023  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains supporting Python modules to distribute with generated types.
The contained support modules are not part of Nunavut, and one should not attempt to import them,
as they may depend on modules that are not available in the local environment.
"""

import pathlib
import typing

from nunavut._utilities import ResourceType, empty_list_support_files, iter_package_resources

__version__ = "1.0.0"
"""Version of the Python support module."""


def list_support_files(
    resource_type: int = ResourceType.ANY.value,
) -> typing.Generator[pathlib.Path, None, None]:
    """
    Get a list of Python support modules embedded in this package.
    """
    if resource_type & ResourceType.SERIALIZATION_SUPPORT.value:
        return iter_package_resources(__name__, ".j2")
    return empty_list_support_files()  # pragma: no cover



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Code generator built on top of pydsdl.

Nunavut uses pydsdl to generate text files using templates. While these
text files are often source code this module could also be used to generate
documentation or data interchange formats like JSON or XML.

The input to the nunavut library is a list of templates and a list of
``pydsdl.CompositeType`` objects. Typical use of this library is simply
invoking the ``nunavut.generate_all`` method.

"""

import sys as _sys
import re as _re

from ._generators import (
    AbstractGenerator,
    basic_language_context_builder_from_args,
    generate_all,
    generate_all_for_language,
    generate_all_from_namespace,
    generate_all_from_namespace_with_generators,
)
from ._namespace import Namespace
from ._utilities import TEMPLATE_SUFFIX, DefaultValue, ResourceSearchPolicy, ResourceType, YesNoDefault
from ._version import __author__, __copyright__, __email__, __license__, __version__
from .jinja import CodeGenerator, DSDLCodeGenerator, SupportGenerator
from .lang import Language, LanguageContext, LanguageContextBuilder, UnsupportedLanguageError
from .lang._config import LanguageConfig

if _sys.version_info[:2] < (3, 10):  # pragma: no cover
    print("Python 3.10 or newer is required", file=_sys.stderr)
    _sys.exit(1)

_version_info_match = _re.match(r"^(\d+)\.(\d+)\.(\d+)", __version__)
if _version_info_match is None:  # pragma: no cover
    raise ValueError(f"Invalid version string: {__version__}")
__version_info__ = tuple(map(int, _version_info_match.groups()))

__all__ = [
    "AbstractGenerator",
    "CodeGenerator",
    "DefaultValue",
    "DSDLCodeGenerator",
    "generate_all",
    "generate_all_for_language",
    "generate_all_from_namespace",
    "generate_all_from_namespace_with_generators",
    "basic_language_context_builder_from_args",
    "Language",
    "LanguageConfig",
    "LanguageContext",
    "LanguageContextBuilder",
    "Namespace",
    "ResourceType",
    "ResourceSearchPolicy",
    "SupportGenerator",
    "TEMPLATE_SUFFIX",
    "UnsupportedLanguageError",
    "YesNoDefault",
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__version__",
    "__version_info__",
]



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Contains template loaders for Nunavut's Jinja2 environment.
"""

import collections
import importlib
import itertools
import logging
from pathlib import Path
from typing import Any, Callable, Deque, Dict, Iterable, List, Mapping, Optional, Tuple, Type, TypeVar, cast

import pydsdl

from nunavut import Namespace as NunavutNamespace
from nunavut._utilities import TEMPLATE_SUFFIX, ResourceSearchPolicy

from .jinja2 import BaseLoader, Environment, FileSystemLoader, PackageLoader, TemplateNotFound

logger = logging.getLogger(__name__)


DEFAULT_TEMPLATE_PATH = "templates"
DEFAULT_SUPPORT_TEMPLATE_PATH = "support"


# +--------------------------------------------------------------------------------------------------------------------+
# | LOADERS : DSDLTemplateLoader
# +--------------------------------------------------------------------------------------------------------------------+


class DSDLTemplateLoader(BaseLoader):
    """
    Nunavut's DSDL template loader is similar to a choice loader with a file-system loader
    first and a package loader as a fallback. The major difference is a DFS is performed
    on the type hierarchy of the type a template is being loaded for. So, for example,
    if no ``StructureType.j2`` template is found then this loader will look for a ``CompositeType.j2``
    and so on.

    :param Optional[NunavutNamespace] namespace: The namespace to use for loading templates. If None then no package
        loader is created.
    :param Optional[List[Path]] templates_dirs: A list of directories to load templates from using a
        :class:`nunavut.jinja.jinja2.FileSystemLoader`. If ``None`` no filesystem loader is created.
    :param bool followlinks: Argument passed on to the :class:`nunavut.jinja.jinja2.FileSystemLoader` instance.
    :param str builtin_template_path: The name of the package under the target language support package to load
        templates from. This is ignored if ``namespace`` is None.
    :param search_policy: If set to "FIND_ALL" then this loader will search using all loaders and will enumerate
                          templates from all loaders. If set to "FIND_FIRST" then the loader will only use the first
                          loader configured for both search and enumeration.
    :param str encoding: The encoding to use when reading templates from the filesystem.
    :param Any kwargs: Arguments forwarded to the :class:`jinja.jinja2.BaseLoader`.
    """

    def __init__(
        self,
        namespace: Optional[NunavutNamespace] = None,
        templates_dirs: Optional[List[Path]] = None,
        followlinks: bool = False,
        builtin_template_path: str = DEFAULT_TEMPLATE_PATH,
        search_policy: ResourceSearchPolicy = ResourceSearchPolicy.FIND_ALL,
        encoding: str = "utf-8",
        **_: Any,
    ):
        super().__init__()
        self._encoding = encoding
        self._type_to_template_lookup_cache: Dict[pydsdl.Any, Path] = dict()

        if templates_dirs is not None:
            for templates_dir_item in templates_dirs:
                if not templates_dir_item.exists():
                    raise ValueError(f"Templates directory {str(templates_dir_item)} did not exist?")
            logger.info("Loading templates from file system at %s", templates_dirs)
            self._fs_loader = FileSystemLoader(
                list(str(d) for d in templates_dirs), followlinks=followlinks, encoding=encoding
            )
        else:
            self._fs_loader = None

        if namespace is not None and (search_policy == ResourceSearchPolicy.FIND_ALL or self._fs_loader is None):
            package_name_for_templates = (
                namespace.get_language_context().get_target_language().get_templates_package_name()
            )
            logger.info("Loading templates from package %s.%s", builtin_template_path, builtin_template_path)
            self._package_loader = PackageLoader(
                package_name_for_templates, package_path=builtin_template_path, encoding=encoding
            )
            self._templates_package_name = f"{package_name_for_templates}.{builtin_template_path}"
        else:
            self._package_loader = None
            self._templates_package_name = ""

    # --[ BaseLoader Overrides ]---------------------------------------------------------------------------------------

    def get_source(self, environment: Environment, template: str) -> Tuple[Any, str, Callable[..., bool]]:
        """
        Override of :meth:`BaseLoader.get_source` that returns the sources of the template from the filesystem loader
        and/or the package loader.
        """
        if self._fs_loader is not None:
            try:
                return cast(
                    Tuple[Any, str, Callable[..., bool]],
                    self._fs_loader.get_source(environment, template),
                )
            except TemplateNotFound:
                if self._package_loader is None:
                    raise
        if self._package_loader is not None:
            return cast(
                Tuple[Any, str, Callable[..., bool]],
                self._package_loader.get_source(environment, template),
            )
        raise TemplateNotFound(template)

    def list_templates(self) -> Iterable[str]:
        """
        Override of :meth:`BaseLoader.list_templates` that returns an aggregate of the filesystem loader and
        package loader templates.

        :return: A list of templates names (i.e. file stems) found by this Generator object.

        .. invisible-code-block: python

            from nunavut.jinja.loaders import DSDLTemplateLoader
            from nunavut._utilities import TEMPLATE_SUFFIX
            from unittest.mock import MagicMock

            mock_namespace = MagicMock()
            mock_get_target_language = mock_namespace.get_language_context.return_value.get_target_language
            mock_get_target_language.return_value.get_templates_package_name.return_value = 'nunavut.lang.c'

            template_loaders = DSDLTemplateLoader(namespace=mock_namespace)

            templates = template_loaders.list_templates()

            structure_type = None

            for template in templates:
                if template == 'StructureType' + TEMPLATE_SUFFIX:
                    structure_type = template

            assert structure_type is not None

        """
        iterables = []
        if self._fs_loader is not None:
            iterables.append(self._filter_template_list_by_suffix(self._fs_loader.list_templates()))
        if self._package_loader is not None:
            iterables.append(self._filter_template_list_by_suffix(self._package_loader.list_templates()))

        return itertools.chain(*iterables)

    # --[ PUBLIC ]-----------------------------------------------------------------------------------------------------

    @property
    def encoding(self) -> str:
        """
        The encoding used to read templates from the filesystem.
        """
        return self._encoding

    def get_templates(self) -> Iterable[Path]:
        """
        Enumerate all templates found in the templates path.
        :data:`~TEMPLATE_SUFFIX` as the suffix for the filename. This method differs from the :class:`BaseLoader`
        override of :meth:`BaseLoader.list_templates` in that it returns paths instead of just file name stems.

        :return: A list of paths to all templates found by this Generator object.

        .. invisible-code-block: python

            from nunavut.jinja.loaders import DSDLTemplateLoader
            from nunavut._utilities import TEMPLATE_SUFFIX
            from unittest.mock import MagicMock

            mock_namespace = MagicMock()
            mock_get_target_language = mock_namespace.get_language_context.return_value.get_target_language
            mock_get_target_language.return_value.get_templates_package_name.return_value = 'nunavut.lang.c'

            template_loaders = DSDLTemplateLoader(namespace=mock_namespace)

            templates = template_loaders.get_templates()

            structure_type = None

            for template in templates:
                if template.stem == 'StructureType':
                    structure_type = template
                assert template.suffix == TEMPLATE_SUFFIX

            assert structure_type is not None
            assert structure_type.suffix == TEMPLATE_SUFFIX
            assert structure_type.exists()

        """
        files = set()
        if self._fs_loader is not None:
            for template_dir in self._fs_loader.searchpath:
                for template in Path(str(template_dir)).glob(f"**/*{TEMPLATE_SUFFIX}"):
                    files.add(template)
        if self._package_loader is not None:
            templates_module = importlib.import_module(self._templates_package_name)
            spec_perhaps = templates_module.__spec__
            file_perhaps: Optional[str] = None
            if spec_perhaps is not None:
                file_perhaps = spec_perhaps.origin
            if file_perhaps is None or file_perhaps == "builtin":
                raise RuntimeError("Unknown template package origin?")
            templates_base_path = Path(file_perhaps).parent
            for t in map(Path, self._filter_template_list_by_suffix(self._package_loader.list_templates())):
                files.add(templates_base_path / t)
        return sorted(files)

    def type_to_template(self, value_type: Type) -> Optional[Path]:
        """
        Given a type object, return a template used to generate code for the type.

        :return: a template or None if no template could be found for the given type.

        .. invisible-code-block: python
            from nunavut.jinja.loaders import DSDLTemplateLoader
            import pydsdl
            from unittest.mock import MagicMock

            mock_namespace = MagicMock()
            mock_get_target_language = mock_namespace.get_language_context.return_value.get_target_language
            mock_get_target_language.return_value.get_templates_package_name.return_value = 'nunavut.lang.py'

            l = DSDLTemplateLoader(namespace=mock_namespace)
            template_name = l.type_to_template(pydsdl.StructureType)

            assert template_name is not None
            assert template_name.name == 'StructureType.j2'

        """
        return self._to_template(self._type_to_template_internal, value_type)

    def index_file_to_template(self, index_file: Path) -> Optional[Path]:
        """
        Given an index file output path, return a template used to render the index.

        :return: a template or None if no template could be found for the given index file.

        .. invisible-code-block: python
            from nunavut.jinja.loaders import DSDLTemplateLoader
            import pydsdl
            from pathlib import Path
            from unittest.mock import MagicMock

            mock_namespace = MagicMock()
            mock_get_target_language = mock_namespace.get_language_context.return_value.get_target_language
            mock_get_target_language.return_value.get_templates_package_name.return_value = 'nunavut.lang.c'

            l = DSDLTemplateLoader(namespace=mock_namespace)
            template_name = l.index_file_to_template(Path("depfile.dep"))

            assert template_name is not None
            assert template_name.name == 'depfile.j2'

        """
        return self._to_template(self._index_file_to_template_internal, index_file)

    # +----------------------------------------------------------------------------------------------------------------+
    # | PRIVATE
    # +----------------------------------------------------------------------------------------------------------------+
    @classmethod
    def _filter_template_list_by_suffix(cls, template_list: Iterable[str]) -> Iterable[str]:
        return filter(lambda x: Path(x).suffix == TEMPLATE_SUFFIX, template_list)

    FromType = TypeVar("FromType")

    def _to_template(
        self,
        converter: Callable[[FromType, Mapping[str, Path]], Optional[Path]],
        value: FromType,
    ) -> Optional[Path]:
        """
        Adapter that uses a provided template resolution function to resolve a template using the loaders configured
        for this instance.
        :param converter: A function that takes a value and a dictionary of templates and returns a template path.
        """
        template_path = None
        if self._fs_loader is not None:
            filtered_templates = map(Path, self._filter_template_list_by_suffix(self._fs_loader.list_templates()))
            template_path = converter(value, dict(map(lambda x: (x.stem, x), filtered_templates)))
        if template_path is None and self._package_loader is not None:
            filtered_templates = map(Path, self._filter_template_list_by_suffix(self._package_loader.list_templates()))
            template_path = converter(value, dict(map(lambda x: (x.stem, x), filtered_templates)))
        return template_path

    def _index_file_to_template_internal(self, index_file: Path, templates: Mapping[str, Path]) -> Optional[Path]:
        try:
            return templates[index_file.stem]
        except KeyError:
            pass
        try:
            return templates["index"]
        except KeyError:
            return None

    def _type_to_template_internal(self, value_type: Type, templates: Mapping[str, Path]) -> Optional[Path]:
        search_queue: Deque[Type] = collections.deque()
        discovered = set()
        search_queue.appendleft(value_type)
        template_path = None

        while len(search_queue) > 0:
            current_search_type = search_queue.pop()
            try:
                template_path = self._type_to_template_lookup_cache[current_search_type]
                break
            except KeyError:
                pass

            try:
                logging.debug(
                    "NunavutTemplateLoader.type_to_template for %s: considering %s...",
                    value_type.__name__,
                    current_search_type.__name__,
                )
                template_path = templates[current_search_type.__name__]
                self._type_to_template_lookup_cache[current_search_type] = template_path
                break
            except KeyError:
                for base_type in current_search_type.__bases__:
                    if base_type != object and base_type not in discovered:
                        search_queue.appendleft(base_type)
                        discovered.add(current_search_type)

        return template_path


# +--------------------------------------------------------------------------------------------------------------------+
# | LOADERS : DSDLSupportTemplateLoader
# +--------------------------------------------------------------------------------------------------------------------+


class DSDLSupportTemplateLoader(DSDLTemplateLoader):
    """
    DSDLTemplateLoader that specializes in loading language support templates.

    :param Any kwargs: Arguments forwarded to the :class:`DSDLTemplateLoader` baseclass.
    """

    def __init__(
        self,
        namespace: NunavutNamespace,
        resource_types: int,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._target_language = namespace.get_language_context().get_target_language()
        self._resource_types = resource_types
        self._support_files_index: Optional[dict[str, Path]] = None

    # --[ BaseLoader Overrides ]---------------------------------------------------------------------------------------

    def get_source(self, environment: Environment, template: str) -> Tuple[Any, str, Callable[..., bool]]:
        """
        Override of :meth:`DSDLTemplateLoader.get_source` that returns the source a the template from the filesystem.
        Nunavut loaders work differently than Jinja2 loaders in that they discover all templates and support files
        at initialization time.
        """
        try:
            return super().get_source(environment, template)
        except TemplateNotFound:
            pass

        support_file = self.find_support_file(template)

        if support_file is None:
            raise TemplateNotFound(template)

        with support_file.open(encoding=self.encoding) as f:
            contents = f.read()

        mtime = support_file.stat().st_mtime

        def is_modified() -> bool:  # pragma: no cover
            try:
                return support_file.stat().st_mtime == mtime
            except OSError:
                return False

        return contents, support_file.as_posix(), is_modified

    def list_templates(self) -> Iterable[str]:
        """
        Override of :meth:`DSDLTemplateLoader.list_templates` that returns an aggregate that loader's templates and
        support file templates.

        :return: A list of templates names (i.e. file names) found by this Generator object.

        .. invisible-code-block: python

            from nunavut.jinja.loaders import DSDLTemplateLoader, DSDLSupportTemplateLoader
            from nunavut._utilities import ResourceType, TEMPLATE_SUFFIX
            from nunavut.lang import LanguageContext, LanguageContextBuilder
            from unittest.mock import MagicMock

            lctx = LanguageContextBuilder().set_target_language("c").create()

            namespace = MagicMock()
            namespace.get_language_context.return_value = lctx

            template_loaders = DSDLSupportTemplateLoader(namespace, ResourceType.SERIALIZATION_SUPPORT.value)

            templates = template_loaders.list_templates()

            serialization_template = None

            for template in templates:
                if template == f"serialization{TEMPLATE_SUFFIX}":
                    serialization_template = template

            assert serialization_template is not None

        """
        return itertools.chain(
            super().list_templates(),
            map(lambda p: p.name, self.get_support_files()),
        )

    # --[ DSDLTemplateLoader Overrides ]------------------------------------------------------------------------------

    def get_templates(self) -> Iterable[Path]:
        """
        Override of :meth:`DSDLTemplateLoader.get_templates` that returns an aggregate that loader's templates and
        support file templates.

        :return: A list of templates paths found by this Generator object.

        .. invisible-code-block: python

            from nunavut.jinja.loaders import DSDLTemplateLoader, DSDLSupportTemplateLoader
            from nunavut._utilities import ResourceType, TEMPLATE_SUFFIX
            from nunavut.lang import LanguageContext, LanguageContextBuilder
            from unittest.mock import MagicMock

            lctx = LanguageContextBuilder().set_target_language("c").create()

            namespace = MagicMock()
            namespace.get_language_context.return_value = lctx

            template_loaders = DSDLSupportTemplateLoader(namespace, ResourceType.SERIALIZATION_SUPPORT.value)

            templates = template_loaders.get_templates()

            serialization_template = None

            for template in templates:
                if template.name == f"serialization{TEMPLATE_SUFFIX}":
                    serialization_template = template

            assert serialization_template is not None

        """
        return itertools.chain(super().get_templates(), self.get_support_files())

    # --[ PUBLIC ]-----------------------------------------------------------------------------------------------------

    def get_support_files(self) -> Iterable[Path]:
        """
        Enumerate all support files found for the target language based on the resource types provided in the
        constructor.
        """
        return self._get_support_file_index().values()

    def find_support_file(self, name: str, default: Optional[Path] = None) -> Optional[Path]:
        """
        Find a support file by name.
        """
        return self._get_support_file_index().get(name, default)

    # --[ PRIVATE ]----------------------------------------------------------------------------------------------------

    def _get_support_file_index(self) -> Dict[str, Path]:
        if self._support_files_index is None:
            self._support_files_index = dict(
                map(lambda x: (x.name, x), self._target_language.get_support_files(self._resource_types))
            )
        return self._support_files_index



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Objects that utilize command-line inputs to run a program using Nunavut.
"""

import argparse
import itertools
import logging
import sys
from pathlib import Path
from typing import Any, Dict, Optional

from .._generators import basic_language_context_builder_from_args, generate_all
from .._utilities import ResourceType
from ..lang import LanguageContext
from .listers import Lister


class StandardArgparseRunner:
    """
    Runner based on Python argparse. This class delegates most of the generation logic to the :func:`generate_all`
    function providing only additional console output on top of that method's functionality.

    :param argparse.Namespace args: The command line arguments.
    """

    def __init__(self, args: argparse.Namespace):
        self._args = args

    @property
    def args(self) -> argparse.Namespace:
        """
        Access to the arguments this object interprets.
        """
        return self._args

    def run(self) -> int:
        """
        Perform actions defined by the arguments this object was created with. This may generate outputs where
        the arguments have requested this action.
        """
        lister_object: Dict[str, Any] = {}
        if self._args.list_configuration:
            lister_object["configuration"] = self.list_configuration(
                basic_language_context_builder_from_args(**vars(self.args)).create()
            )
            if len(sys.argv) > 0:
                lister_object["cmdline"] = sys.argv

        result = generate_all(**vars(self.args))

        if self._args.list_inputs:
            input_dsdl = {str(p) for p in set(result.template_files)}
            for _, target_data in result.generator_targets.items():
                input_dsdl.add(str(target_data.definition.source_file_path.resolve()))
                input_dsdl.update({str(d.source_file_path.resolve()) for d in target_data.input_types})
            lister_object["inputs"] = list(input_dsdl)

        if self._args.list_outputs:
            file_iterators = []
            if self._args.resource_types != ResourceType.NONE.value:
                file_iterators.append(result.support_files)
            if (self._args.resource_types & ResourceType.ONLY.value) == 0:
                file_iterators.append(result.generated_files)
            lister_object["outputs"] = [str(p.resolve()) for p in itertools.chain(*file_iterators)]

        Lister.get_lister(self._args.list_format, self._args.list_to_file).list(lister_object)

        return 0

    def list_configuration(self, lctx: LanguageContext) -> Dict[str, Any]:
        """
        List the configuration of the language context to an object.
        """

        config: Dict[str, Any] = {}
        config["target_language"] = lctx.get_target_language().name
        config["sections"] = lctx.config.sections()

        return config


# --[ MAIN ]-----------------------------------------------------------------------------------------------------------
def main(command_line_args: Optional[Any] = None) -> int:
    """
    Main entry point for command-line scripts.
    """
    from multiprocessing import freeze_support  # pylint: disable=import-outside-toplevel

    freeze_support()

    from . import _make_parser  # pylint: disable=import-outside-toplevel
    from .parsers import NunavutArgumentParser  # pylint: disable=import-outside-toplevel

    #
    # Parse the command-line arguments.
    #
    parser = _make_parser(NunavutArgumentParser)

    try:
        import argcomplete  # pylint: disable=import-outside-toplevel

        argcomplete.autocomplete(parser)
    except ImportError:
        logging.debug("argcomplete not installed, skipping autocomplete")

    args = parser.parse_args(args=command_line_args)

    #
    # Setup Python logging.
    #
    fmt = "%(message)s"
    level = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}.get(args.verbose or 0, logging.DEBUG)
    logging.basicConfig(stream=sys.stderr, level=level, format=fmt)

    logging.info("Running %s using sys.prefix: %s", Path(__file__).name, sys.prefix)

    return StandardArgparseRunner(args).run()



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Objects that write debug/accounting information about code generation inputs, outputs, and configuration
to various outputs.
"""

import abc
import csv
import json
import sys
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Type

from nunavut._utilities import DefaultValue


class ConfigJSONEncoder(json.JSONEncoder):
    """
    A JSON encoder that can handle Nunavut configuration and pydsdl objects.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, Path):
            return str(o)
        if isinstance(o, DefaultValue):
            return o.value
        return super().default(o)


class Lister(abc.ABC):
    """
    Abstract base class for listers
    """

    DefaultEncoding = "utf-8"

    @classmethod
    def get_lister(cls, list_format: str, list_file: Optional[Path]) -> "Lister":
        """
        Get a lister object based on the given format and file.
        """
        if list_format in ("json", "json-pretty"):
            return JsonLister(list_file=list_file, pretty=list_format == "json-pretty")
        elif list_format in ("csv", "scsv"):
            return ValueSeparatedLister(separator=";" if list_format == "scsv" else ",", list_file=list_file)
        else:
            raise ValueError(f"Unsupported list format: {list_format}")  # pragma: no cover

    @abc.abstractmethod
    def list(self, list_object: Dict[str, Any]) -> None:
        """
        List the given object to the concrete lister's output(s).
        """


class JsonLister(Lister):
    """
    Dumps the given list object as JSON to stdout and optionally to a file.

    .. invisible-code-block: python

        import pathlib
        import json
        from nunavut.cli.listers import JsonLister

        list_file_path = gen_paths_for_module.out_dir / pathlib.Path("lister-test.json")
        list_data = {
            "inputs": ["input1", "input2"],
            "outputs": ["output1", "output2"]
        }

        JsonLister(list_file=list_file_path).list(list_data)

        with list_file_path.open("r", encoding=Lister.DefaultEncoding) as list_file:
            json_data = json.load(list_file)
            assert "inputs" in json_data
            assert "outputs" in json_data
            found_data = set(json_data["inputs"] + json_data["outputs"])

        assert "input1" in found_data
        assert "input2" in found_data
        assert "output1" in found_data
        assert "output2" in found_data

    """

    def __init__(
        self,
        json_encoder: Type[json.JSONEncoder] = ConfigJSONEncoder,
        pretty: bool = False,
        list_file: Optional[Path] = None,
    ):
        self._json_encoder = json_encoder
        self._list_file = list_file
        self._pretty = pretty

    def list(self, list_object: Dict[str, Any]) -> None:
        """
        List the given object to the JSON file and optionally to a file.
        """
        if self._pretty:
            indent = 2
        else:
            indent = None
        json.dump(list_object, sys.stdout, ensure_ascii=False, indent=indent, cls=self._json_encoder)
        if self._list_file is not None:
            with self._list_file.open("w", encoding=self.DefaultEncoding) as list_file:
                json.dump(list_object, list_file, ensure_ascii=False, indent=indent, cls=self._json_encoder)


class ValueSeparatedLister(Lister):
    """
    Dumps the given list object to stdout and optionally to a file with a given separator.

    .. invisible-code-block: python

        import pathlib
        import csv
        from nunavut.cli.listers import ValueSeparatedLister

        list_file_path = gen_paths_for_module.out_dir / pathlib.Path("lister-test.csv")
        list_data = {
            "inputs": ["input1", "input2"],
            "outputs": ["output1", "output2"]
        }

        ValueSeparatedLister(separator=",", list_file=list_file_path).list(list_data)

        found_data = set()

        with list_file_path.open("r", encoding=Lister.DefaultEncoding) as list_file:
            csv_reader = csv.reader(list_file, delimiter=',')
            for row in csv_reader:
                found_data.update(row)

        assert "input1" in found_data
        assert "input2" in found_data
        assert "output1" in found_data
        assert "output2" in found_data

    """

    def __init__(self, separator: str, list_file: Optional[Path] = None):
        self._list_file = list_file
        self._sep = separator

    def list(self, list_object: Dict[str, Any]) -> None:
        self.stdout_lister(list_object)

        if self._list_file is not None:
            with self._list_file.open("w", encoding=self.DefaultEncoding) as list_file:
                csv_writer = csv.writer(list_file, delimiter=self._sep)
                for key, value in list_object.items():
                    csv_writer.writerow([key] + value)

    def stdout_lister(self, list_object: Dict[str, Any]) -> None:
        """
        The output dialect to stdout is not supported by the CSV implementation in python. Use this instead.
        """

        def _write_row(row: Iterable[str], sep: str, end: str) -> None:
            first = True
            for cell in row:
                if first:
                    first = False
                else:
                    sys.stdout.write(sep)
                sys.stdout.write(cell)
            if not first:
                sys.stdout.write(end)

        had_inputs = False
        has_outputs = "outputs" in list_object and len(list_object["outputs"]) > 0
        if "inputs" in list_object and len(list_object["inputs"]) > 0:
            had_inputs = True
            _write_row(list_object["inputs"], sep=self._sep, end=(self._sep if has_outputs else ""))
        if has_outputs:
            if had_inputs:
                sys.stdout.write(self._sep)
            _write_row(list_object["outputs"], sep=self._sep, end="")



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Command-line parsers for the Nunavut command-line interface.
"""

import argparse
import itertools
import os
import re
import sys
import textwrap
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

from nunavut._postprocessors import (
    ExternalProgramEditInPlace,
    LimitEmptyLines,
    PostProcessor,
    SetFileMode,
    TrimTrailingWhitespace,
)
from nunavut._utilities import DefaultValue, QuaternaryLogic, ResourceSearchPolicy, ResourceType, YesNoDefault


class NunavutArgumentParser(argparse.ArgumentParser):
    """
    Specialization of argparse.ArgumentParser to encapsulate inter-argument rules, aggregate path arguments, and
    combine language options. The output of this parser can be fed directly into the :func:`nunavut.generate_all`
    function.

    Arguments Added
    ---------------

    - **target_files**
        A list of target files to process with colon syntax resolved.

    - **root_namespace_directories_or_names**
        A list of additional directories to search for DSDL files. This is a combination of the
        lookup_dir, results after parsing target file colon syntax, and paths from supported
        environment variables.

    - **language_options**
        A dictionary of options to pass to a language context builder.

    - **resource_types**
        Bitmask of ResourceType values to generate based on omit_serialization_support and generate_support arguments.

    - **post_processors**
        A list of post processors to run on generated files.

    - **search_policy**
        The ResourceSearchPolicy to use when searching for templates.

    Arguments Removed
    -----------------

    - **generate_support**
        The original argument is replaced with resource_types bitmask.

    - **target_files_or_root_namespace**
        The original argument is replaced with target_files and root_namespace_directories_or_names.

    - **lookup_dir**
        The original argument is replaced with root_namespace_directories_or_names.

    - **fallback_to_builtin_templates**
        Transformed into search_policy

    - **pp_trim_trailing_whitespace**
        The original arguments are replaced with post_processors.

    - **pp_max_emptylines**
        The original arguments are replaced with post_processors.

    - **pp_run_program**
        The original arguments are replaced with post_processors.

    - **pp_run_program_arg**
        The original arguments are replaced with post_processors.

    Arguments Modified
    ------------------

    - **verbose**
        If not provided, defaults to 0.

    - **list_outputs**
        If provided, sets dry_run to True.

    - **list_inputs**
        If provided, sets dry_run to True.

    - **generate_namespace_types**
        Converts the argument to a YesNoDefault enum.

    """

    DSDL_FILE_SUFFIXES = (".uavcan", ".dsdl")
    MAX_LOG_ENTRY_LENGTH = 1000

    # --[ OVERRIDE ]--------------------------------------------------------------------------------------------------
    def parse_known_args(self, args=None, namespace=None):  # type: ignore
        parsed_args, argv = super().parse_known_args(args, namespace)
        self._post_process_args(parsed_args)
        return (parsed_args, argv)

    # --[ PRIVATE ]---------------------------------------------------------------------------------------------------

    @staticmethod
    def _sanitize_log_message(message: str) -> str:
        """
        Sanitize a message before logging by:
        1. Removing control characters
        2. Replacing newlines and tabs with spaces
        3. Limiting length

        :param: Raw message to sanitize

        :return: Sanitized message safe for logging

         .. invisible-code-block: python

            from nunavut.cli.parsers import NunavutArgumentParser

            # Control characters
            message = "This is a test message with control characters \\x00\\x01."
            sanitized_message = "This is a test message with control characters ."
            assert sanitized_message == NunavutArgumentParser._sanitize_log_message(message)

            # Newlines and tabs
            message = "This is a\\r\\ntest\\tmessage with newlines and tabs.\\n\\t"
            sanitized_message = "This is a test message with newlines and tabs."
            assert sanitized_message == NunavutArgumentParser._sanitize_log_message(message)

            # Multiple spaces
            message = "This is a test message    with    multiple    spaces."
            sanitized_message = "This is a test message with multiple spaces."
            assert sanitized_message == NunavutArgumentParser._sanitize_log_message(message)

            # Stripping whitespace
            message = "   This is a test message with leading and trailing whitespace.   "
            sanitized_message = "This is a test message with leading and trailing whitespace."
            assert sanitized_message == NunavutArgumentParser._sanitize_log_message(message)

            # Length limit
            message_preamble = "This is a test message"
            message = message_preamble + "a" * NunavutArgumentParser.MAX_LOG_ENTRY_LENGTH
            sanitized_message = (
                message_preamble +
                "a" * (NunavutArgumentParser.MAX_LOG_ENTRY_LENGTH - len(message_preamble)) + "... (truncated)"
            )
            assert sanitized_message == NunavutArgumentParser._sanitize_log_message(message)

        """
        # Remove control characters except whitespace
        message = "".join(char for char in message if char.isprintable() or char.isspace())

        # Replace newlines and tabs with spaces
        message = re.sub(r"[\n\r\t]+", " ", message)

        # Collapse multiple spaces
        message = re.sub(r"\s+", " ", message)

        # Trim whitespace
        message = message.strip()

        # Limit length to prevent log flooding
        if len(message) > NunavutArgumentParser.MAX_LOG_ENTRY_LENGTH:
            message = message[: NunavutArgumentParser.MAX_LOG_ENTRY_LENGTH] + "... (truncated)"

        return message

    def _post_process_log(self, args: argparse.Namespace, message: str) -> None:
        """
        Print a message to the log.
        """
        if args.verbose <= 0:
            return
        sanitary_message = self._sanitize_log_message(message)
        self._print_message(sanitary_message, sys.stdout)

    def _post_process_args(self, args: argparse.Namespace) -> None:
        """
        Applies rules between different arguments and handles other special cases.
        """

        if args.verbose is None:
            args.verbose = 0

        if args.list_outputs:
            args.dry_run = True

        if args.list_inputs:
            args.dry_run = True

        if args.fallback_to_builtin_templates:
            args.search_policy = ResourceSearchPolicy.FIND_ALL
        else:
            args.search_policy = ResourceSearchPolicy.FIND_FIRST

        del args.fallback_to_builtin_templates

        # Convert omit_serialization_support and generate_support to resource_types bitmask.
        args.resource_types = self._create_resource_types_bitmask(args)

        # Find all possible path specifications and combine into two collections: root_namespace_directories_or_names
        # and target_files.
        # TARGETS
        root_namespace_directories_or_names, target_files = self._parse_target_paths(
            args.target_files_or_root_namespace, True, error_if_folder=args.no_target_namespaces
        )
        del args.target_files_or_root_namespace

        # LOOKUP PATHS
        lookup_paths = args.lookup_dir
        del args.lookup_dir

        root_namespace_directories_or_names_from_lookups, target_files_from_lookups = self._parse_target_paths(
            lookup_paths,
            False,
        )

        root_namespace_directories_or_names.update(root_namespace_directories_or_names_from_lookups)
        target_files.update(target_files_from_lookups)

        # PATHS FROM ENVIRONMENT

        root_namespace_directories_or_names_from_environment, target_files_from_environment = self._parse_target_paths(
            [str(env_path) for env_path in self._lookup_paths_from_environment(args)],
            False,
        )

        root_namespace_directories_or_names.update(root_namespace_directories_or_names_from_environment)
        target_files.update(target_files_from_environment)

        # Store the combined paths in args
        args.target_files = list(target_files)
        args.root_namespace_directories_or_names = list(root_namespace_directories_or_names)

        # Create a dictionary of language options.
        args.language_options = self._create_language_options(args)

        # Create a list of post processors.
        args.post_processors = self._create_post_processors(args)

        # Generator arguments
        args.generate_namespace_types = YesNoDefault.YES if args.generate_namespace_types else YesNoDefault.DEFAULT

        # Can't list configuration as csv. Has to be a structured return format.
        if args.list_configuration and args.list_format in ("scsv", "csv"):
            self.error(textwrap.dedent(f"""

            --list-format {args.list_format} is not supported for --list-configuration. Use a structured format like
            --list-format json to list configuration information.

            """))

    def _parse_target_paths(
        self, target_files_or_root_namespace: Optional[List[str]], greedy: bool, error_if_folder: bool = False
    ) -> Tuple[Set[Path], Set[Path]]:
        """
        Parse the target paths from the command line arguments.

        :param target_files_or_root_namespace: The target files or root namespace directories.
        :param greedy: If True, and if a path is a simple folder (i.e. no colon-syntax), then the path will
            be considered a root path and all DSDL files found in the folder, using globular magic, will be considered
            target files.

        :return: A list of root paths (folders) and a list of target paths (files).

        .. invisible-code-block: python

            from nunavut.cli.parsers import NunavutArgumentParser
            from pytest import raises
            from pathlib import Path

            real_root = Path().cwd().as_posix()

            parser = NunavutArgumentParser()
            # Happy path
            root_paths, target_files = parser._parse_target_paths(
                [
                    f"{real_root}one/to/root",
                    f"{real_root}two/to/file.dsdl",
                    "three/path/four:to/file.dsdl",
                    f"{real_root}five/path/six:to/file.dsdl",
                    "seven/path/eight\\\\:to/file.dsdl",
                    f"{real_root}nine/path/ten/:to/file.dsdl",
                ],
                False,
            )
            print(root_paths, target_files)
            assert len(root_paths) == 4
            assert len(target_files) == 5

            assert Path(f"{real_root}one/to/root") in root_paths
            assert Path("three/path/four") in root_paths
            assert Path(f"{real_root}five/path/six") in root_paths
            assert Path(f"{real_root}nine/path/ten") in root_paths

            assert Path(f"{real_root}two/to/file.dsdl") in target_files
            assert Path("four/to/file.dsdl") in target_files
            assert Path("six/to/file.dsdl") in target_files
            assert Path("seven/path/eight\\\\:to/file.dsdl") in target_files
            assert Path("ten/to/file.dsdl") in target_files

            # Happy path: default root path, no dependencies
            default_root_paths, default_target_files = parser._parse_target_paths([""], False)
            assert len(default_root_paths) == 1
            assert len(default_target_files) == 0
            assert default_root_paths.pop() == Path().cwd()

            # Happy path: single target file
            single_target_file_root_paths, single_target_file_target_files = parser._parse_target_paths(
                [f"{real_root}one/to/file.dsdl"], True
            )
            assert len(single_target_file_root_paths) == 0
            assert len(single_target_file_target_files) == 1
            assert single_target_file_target_files.pop() == Path(f"{real_root}one/to/file.dsdl")

            # errors: multiple colons
            with raises(SystemExit):
                parser._parse_target_paths(["one:two:three"], False)

            # errors: leading slash
            with raises(SystemExit):
                parser._parse_target_paths([f"path/to:{Path.cwd().anchor}root/to/file.dsdl"], False)

        """

        def _parse_lookup_dir(lookup_dir: str) -> Tuple[Optional[Path], Optional[Path]]:

            lookup_path = Path(lookup_dir)
            if lookup_path.is_absolute():
                relative_lookup_dir = lookup_path.relative_to(lookup_path.anchor).as_posix()
            else:
                relative_lookup_dir = lookup_dir
            split_path = re.split(r"(?<!\\):", relative_lookup_dir)
            if len(split_path) > 2:
                self.error(f"Invalid lookup path (too many colons) > {lookup_dir}")
            if lookup_path.is_absolute():
                root_path = Path(lookup_path.anchor, split_path[0])
            else:
                root_path = Path(split_path[0])
            if len(split_path) == 2:
                return root_path, Path(root_path.stem, split_path[1])
            elif root_path.suffix in self.DSDL_FILE_SUFFIXES:
                return None, root_path
            else:
                return root_path, None

        if target_files_or_root_namespace is None:
            return set(), set()

        root_paths = set()
        target_files = set()
        for target_path in target_files_or_root_namespace:
            root_dir, target_file_maybe = _parse_lookup_dir(target_path)
            if root_dir is not None:
                if target_file_maybe is None:
                    if error_if_folder:
                        self.error(f"Root directory cannot be a folder (--error-if-folder) > {root_dir}")
                    root_dir_resolved = root_dir.expanduser().resolve()
                    if root_dir_resolved.parent == root_dir_resolved:
                        raise RuntimeError(f"Refusing to search root directory > {root_dir_resolved}")
                    # only a path so we'll use globular magic to find all the DSDL files in the directory
                    root_paths.add(root_dir_resolved)
                    if greedy:
                        for dsdl_suffix in self.DSDL_FILE_SUFFIXES:
                            target_files.update(set(root_dir_resolved.glob(f"**/*{dsdl_suffix}")))
                else:
                    root_paths.add(root_dir)
            if target_file_maybe is not None:
                if root_dir is not None and target_file_maybe.is_absolute():
                    self.exit(
                        1,
                        f"Target file path cannot be absolute when using colon syntax > {root_dir}:{target_file_maybe}",
                    )
                target_files.add(target_file_maybe)

        return root_paths, target_files

    def _lookup_paths_from_environment(self, args: argparse.Namespace) -> Set[Path]:
        """
        Parse supported environment variables
        """

        def _extra_includes_from_env(env_var_name: str) -> List[Path]:
            try:
                return [Path(extra) for extra in os.environ[env_var_name].split(os.pathsep)]
            except KeyError:
                return []

        extra_includes: List[Path] = []

        dsdl_include_path = _extra_includes_from_env("DSDL_INCLUDE_PATH")

        if len(dsdl_include_path) > 0:
            self._post_process_log(args, f"Extra includes from DSDL_INCLUDE_PATH: {dsdl_include_path}")

        cyphal_root_paths = _extra_includes_from_env("CYPHAL_PATH")
        cyphal_paths = [
            c
            for c in itertools.chain.from_iterable(map(lambda cyphal_path: cyphal_path.glob("*"), cyphal_root_paths))
            if c.is_dir()
        ]

        if len(cyphal_paths) > 0:
            self._post_process_log(args, f"Extra includes from CYPHAL_PATH: {cyphal_paths}")

        extra_includes += sorted(itertools.chain(dsdl_include_path, cyphal_paths))

        return set(extra_includes)

    @classmethod
    def _create_language_options(cls, args: argparse.Namespace) -> Dict[str, Any]:
        """
        Group all language options into a dictionary.
        """
        language_options = {}
        if args.target_endianness is not None:
            language_options["target_endianness"] = args.target_endianness

        language_options["omit_float_serialization_support"] = (
            True if args.omit_float_serialization_support else DefaultValue(False)
        )

        language_options["omit_serialization_support"] = (
            True if args.omit_serialization_support else DefaultValue(False)
        )

        language_options["enable_serialization_asserts"] = (
            True if args.enable_serialization_asserts else DefaultValue(False)
        )

        language_options["enable_override_variable_array_capacity"] = (
            True if args.enable_override_variable_array_capacity else DefaultValue(False)
        )

        if args.language_standard is not None:
            language_options["std"] = args.language_standard

        return language_options

    @classmethod
    def _create_post_processors(cls, args: argparse.Namespace) -> List[PostProcessor]:
        """
        A, possibly empty, list of post processors to run based on provided arguments.
        """

        def _build_ext_program_postprocessor_args(program: str) -> List[str]:
            """
            Build an array of arguments for the program.
            """
            subprocess_args = [program]
            if hasattr(args, "pp_run_program_arg") and args.pp_run_program_arg is not None:
                for program_arg in args.pp_run_program_arg:
                    subprocess_args.append(program_arg)
            return subprocess_args

        post_processors: List[PostProcessor] = []
        if args.pp_trim_trailing_whitespace:
            post_processors.append(TrimTrailingWhitespace())
            del args.pp_trim_trailing_whitespace
        if hasattr(args, "pp_max_emptylines") and args.pp_max_emptylines is not None:
            post_processors.append(LimitEmptyLines(args.pp_max_emptylines))
            del args.pp_max_emptylines
        if hasattr(args, "pp_run_program") and args.pp_run_program is not None:
            post_processors.append(
                ExternalProgramEditInPlace(_build_ext_program_postprocessor_args(args.pp_run_program))
            )
            del args.pp_run_program
            del args.pp_run_program_arg

        post_processors.append(SetFileMode(args.file_mode))

        return post_processors

    @classmethod
    def _create_resource_types_bitmask(cls, args: argparse.Namespace) -> int:
        """
        Create a bitmask of ResourceType values based on the provided arguments. This method utilizes generate_support
        and omit_serialization_support arguments to determine the bitmask. It also removes the generate_support argument
        from the namespace.

        :param args: The parsed arguments.

        .. invisible-code-block: python

            from nunavut.cli.parsers import NunavutArgumentParser, QuaternaryLogic, ResourceType

            def _test_create_resource_types_bitmask(generate_support, omit_serialization_support, expected):
                namespace = argparse.Namespace(
                    generate_support=generate_support,
                    omit_serialization_support=omit_serialization_support
                )

                resource_types = NunavutArgumentParser._create_resource_types_bitmask(namespace)
                assert resource_types == expected
                assert "generate_support" not in namespace


            _test_create_resource_types_bitmask("only", False, (
                ResourceType.ONLY.value | ResourceType.SERIALIZATION_SUPPORT.value | ResourceType.TYPE_SUPPORT.value)
            )
            _test_create_resource_types_bitmask("never", True, ResourceType.NONE.value)
            _test_create_resource_types_bitmask("as-needed", False, (
                ResourceType.ANY.value | ResourceType.SERIALIZATION_SUPPORT.value | ResourceType.TYPE_SUPPORT.value)
            )
            _test_create_resource_types_bitmask("always", False, (
                ResourceType.ONLY.value | ResourceType.SERIALIZATION_SUPPORT.value | ResourceType.TYPE_SUPPORT.value)
            )

        """
        generate_support = QuaternaryLogic.from_en_us(args.generate_support)
        del args.generate_support

        resource_types = ResourceType.ANY.value

        if generate_support is QuaternaryLogic.ALWAYS_FALSE:
            resource_types &= ~ResourceType.TYPE_SUPPORT.value
            resource_types &= ~ResourceType.SERIALIZATION_SUPPORT.value

        elif generate_support is QuaternaryLogic.TRUE_IF:
            resource_types = (
                ResourceType.ANY.value | ResourceType.SERIALIZATION_SUPPORT.value | ResourceType.TYPE_SUPPORT.value
            )
        elif generate_support is QuaternaryLogic.ALWAYS_TRUE or generate_support is QuaternaryLogic.TRUE_UNLESS:
            resource_types = (
                ResourceType.ONLY.value | ResourceType.SERIALIZATION_SUPPORT.value | ResourceType.TYPE_SUPPORT.value
            )

        if args.omit_serialization_support:
            resource_types &= ~ResourceType.SERIALIZATION_SUPPORT.value

        return resource_types



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
# cSpell: words loopcontrols
#
"""
Jinja environment for Nunavut code generation.
"""

import datetime
import inspect
import logging
import platform
import sys
import types
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    ItemsView,
    KeysView,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
    ValuesView,
    cast,
)

from nunavut._templates import LanguageEnvironment
from nunavut.lang import Language, LanguageClassLoader, LanguageContext

from .extensions import JinjaAssert, UseQuery
from .jinja2 import BaseLoader, Environment, StrictUndefined, select_autoescape
from .jinja2.ext import Extension
from .jinja2.ext import do as jinja_do
from .jinja2.ext import loopcontrols
from .jinja2.filters import FILTERS as JINJA2_FILTERS

logger = logging.getLogger(__name__)

# +---------------------------------------------------------------------------+
# | JINJA : LanguageTemplateNamespace
# +---------------------------------------------------------------------------+


class LanguageTemplateNamespace:
    """
    Generic namespace object used to create reserved namespaces in the global environment.

    .. invisible-code-block: python

        from nunavut.jinja.environment import LanguageTemplateNamespace

    .. code-block:: python

        ns = LanguageTemplateNamespace()

        # any property can be set at any time.
        ns.foo = 'foo'
        assert ns.foo == 'foo'

        # repr of the ns enables cloning using exec
        exec('ns2={}'.format(repr(ns)))
        assert ns2.foo == 'foo'

        # clones will be equal
        assert ns2 == ns

        # but not the same object
        assert ns2 is not ns

    In addition to the namespace behavior this object exposed some dictionary-like methods:

    .. code-block:: python

        ns = LanguageTemplateNamespace()
        ns.update({'foo':'bar'})

        assert ns.foo == 'bar'

    .. invisible-code-block: python

        ns = LanguageTemplateNamespace(one='one', two='two')
        assert 'one' == ns.one
        assert 'two' == ns.two

    """

    def __init__(self, **kwargs: Any):
        for name, value in kwargs.items():
            setattr(self, name, value)

    def __repr__(self) -> str:
        type_name = type(self).__name__
        arg_strings = []
        star_args = {}
        for name, value in self._get_kwargs():
            if name.isidentifier():
                arg_strings.append(f"{name}={repr(value)}")
            else:
                star_args[name] = value
        if star_args:
            arg_strings.append(f"**{repr(star_args)}")
        return f"{type_name}({','.join(arg_strings)})"

    def _get_kwargs(self) -> List[Any]:
        return list(self.__dict__.items())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, LanguageTemplateNamespace):
            return NotImplemented
        return vars(self) == vars(other)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def update(self, update_from: Mapping[str, Any]) -> None:
        """
        update the namespace with the given values.
        """
        for key, value in update_from.items():
            setattr(self, key, value)

    def items(self) -> ItemsView[str, Any]:
        """
        The items in the namespace.
        """
        return self.__dict__.items()

    def keys(self) -> KeysView[Any]:
        """
        The values in the namespace.
        """
        return self.__dict__.keys()

    def values(self) -> ValuesView[Any]:
        """
        The values in the namespace.
        """
        return self.__dict__.values()


# +---------------------------------------------------------------------------+
# | JINJA : CodeGenEnvironment
# +---------------------------------------------------------------------------+
class CodeGenEnvironmentBuilder:
    """
    Builder class for creating a CodeGenEnvironment object for code generation.

    :param BaseLoader loader: The loader used to load templates.
    :param LanguageContext lctx: The language context used for code generation.
    """

    DEFAULT_JINJA_EXTENSIONS = [jinja_do, loopcontrols, JinjaAssert, UseQuery]

    def __init__(self, loader: BaseLoader) -> None:
        self._loader = loader
        self._trim_blocks = False
        self._lstrip_blocks = False
        self._additional_filters: Optional[Dict[str, Callable]] = None
        self._additional_tests: Optional[Dict[str, Callable]] = None
        self._additional_globals: Optional[Dict[str, Any]] = None
        self._extensions = self.DEFAULT_JINJA_EXTENSIONS[:]
        self._allow_filter_test_or_use_query_overwrite = False
        self._embed_auditing_info = False

    @property
    def loader(self) -> BaseLoader:
        """
        The loader.

        :return: The loader.
        :rtype: BaseLoader
        """
        return self._loader

    def set_trim_blocks(self, trim_blocks: bool) -> "CodeGenEnvironmentBuilder":
        """
        Set the trim blocks flag.

        :param bool trim_blocks: The trim blocks flag.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        self._trim_blocks = trim_blocks
        return self

    def set_lstrip_blocks(self, lstrip_blocks: bool) -> "CodeGenEnvironmentBuilder":
        """
        Set the lstrip blocks flag.

        :param bool lstrip_blocks: The lstrip blocks flag.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        self._lstrip_blocks = lstrip_blocks
        return self

    def add_filters(self, **additional_filters: Callable) -> "CodeGenEnvironmentBuilder":
        """
        Add filters to the created environment.

        :param Dict[str, Callable] additional_filters: The additional filters.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        if self._additional_filters is None:
            self._additional_filters = additional_filters
        else:
            self._additional_filters.update(additional_filters)
        return self

    def add_tests(self, **additional_tests: Callable) -> "CodeGenEnvironmentBuilder":
        """
        Add tests to the created environment.

        :param Dict[str, Callable] additional_tests: The additional tests.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        if self._additional_tests is None:
            self._additional_tests = additional_tests
        else:
            self._additional_tests.update(additional_tests)
        return self

    def add_globals(self, **additional_globals: Any) -> "CodeGenEnvironmentBuilder":
        """
        Add globals so the created environment.

        :param Dict[str, Any] additional_globals: The additional globals.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        if self._additional_globals is None:
            self._additional_globals = additional_globals
        else:
            self._additional_globals.update(additional_globals)
        return self

    def set_extensions(self, *extensions: Extension) -> "CodeGenEnvironmentBuilder":
        """
        Set the extensions.

        :param List[Extension] extensions: The extensions.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        self._extensions = list(extensions)
        return self

    def set_allow_filter_test_or_use_query_overwrite(
        self, allow_filter_test_or_use_query_overwrite: bool
    ) -> "CodeGenEnvironmentBuilder":
        """
        Allow overwriting of built-in filters, tests, or use queries.

        :param bool allow_filter_test_or_use_query_overwrite: Allow overwrite of built-ins.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        self._allow_filter_test_or_use_query_overwrite = allow_filter_test_or_use_query_overwrite
        return self

    def set_embed_auditing_info(self, embed_auditing_info: bool) -> "CodeGenEnvironmentBuilder":
        """
        Set whether to embed auditing information in generated files.

        :param bool embed_auditing_info: Whether to embed auditing information.
        :return: The CodeGenEnvironmentBuilder object.
        :rtype: CodeGenEnvironmentBuilder
        """
        self._embed_auditing_info = embed_auditing_info
        return self

    def create(self, lctx: LanguageContext) -> "CodeGenEnvironment":
        """
        Create a CodeGenEnvironment object.

        :return: A CodeGenEnvironment object.
        :rtype: CodeGenEnvironment
        """
        env = CodeGenEnvironment(
            self.loader,
            trim_blocks=self._trim_blocks,
            lstrip_blocks=self._lstrip_blocks,
            additional_filters=self._additional_filters,
            additional_tests=self._additional_tests,
            additional_globals=self._additional_globals,
            extensions=self._extensions,
            allow_filter_test_or_use_query_overwrite=self._allow_filter_test_or_use_query_overwrite,
            embed_auditing_info=self._embed_auditing_info,
        )
        env.set_language_context(lctx)
        return env


# +---------------------------------------------------------------------------+


class CodeGenEnvironment(Environment):
    """
    Jinja Environment optimized for compile-time generation of source code
    (i.e. as opposed to dynamically generating webpages).

    Do not instanceate directly. Use the :class:`CodeGenEnvironmentBuilder` to create an instance.

    .. invisible-code-block: python

        from nunavut.lang import LanguageContext, LanguageContextBuilder
        from nunavut.lang._language import Language
        from nunavut.jinja import CodeGenEnvironmentBuilder
        from nunavut.jinja.jinja2 import DictLoader

        lctx = LanguageContextBuilder().create()

    .. code-block:: python

        template = 'Hello World'

        e = CodeGenEnvironmentBuilder(DictLoader({'test': template})).create(lctx)
        assert 'Hello World' ==  e.get_template('test').render()

    .. warning::
        The :attr:`RESERVED_GLOBAL_NAMESPACES` and :attr:`RESERVED_GLOBAL_NAMES` collections
        contain names in the global namespace reserved by this environment. Attempting to override one
        of these reserved names will cause the constructor to raise an error.

    .. code-block:: python

        try:
            (
                CodeGenEnvironmentBuilder(DictLoader({'test': template}))
                .add_globals(ln='bad_ln')
                .create(lctx)
            )
            assert False
        except RuntimeError:
            pass

    Other safe-guards include checks that Jinja built-ins aren't accidentally overridden...

    .. code-block:: python

        try:
            (
                CodeGenEnvironmentBuilder(DictLoader({'test': template}))
                .add_filters(indent=lambda x: x)
                .create(lctx)
            )
            assert False
        except RuntimeError:
            pass

        # You can allow overwrite of built-ins using the ``allow_filter_test_or_use_query_overwrite``
        # argument.
        e = (
                CodeGenEnvironmentBuilder(DictLoader({'test': template}))
                .add_filters(indent=lambda x: x)
                .set_allow_filter_test_or_use_query_overwrite(True)
                .create(lctx)
            )
        assert 'foo' == e.filters['indent']('foo')

    ...or that user-defined filters or redefined.

    .. code-block:: python

        class MyFilters:

            @staticmethod
            def filter_misnamed(name: str) -> str:
                return name

        e = (
                CodeGenEnvironmentBuilder(DictLoader({'test': template}))
                .add_filters(filter_misnamed=lambda x: x)
                .create(lctx)
            )

        try:
            e.add_conventional_methods_to_environment(MyFilters())
            assert False
        except RuntimeError:
            pass

    .. note:: Maintainer's Note
        This class should remain DSDL agnostic. It is, theoretically, applicable using Jinja with any compiler front-end
        input although, in practice, it will only ever be used with pydsdl AST.
        Pydsdl-specific logic should live in the CodeGenerator (:class:`nunavut.jinja.DSDLCodeGenerator`).

    """

    RESERVED_GLOBAL_NAMESPACES = {"ln", "options", "uses_queries", "nunavut"}

    RESERVED_GLOBAL_NAMES = {"now_utc"}

    def __init__(
        self,
        loader: BaseLoader,
        trim_blocks: bool,
        lstrip_blocks: bool,
        additional_filters: Optional[Dict[str, Callable]],
        additional_tests: Optional[Dict[str, Callable]],
        additional_globals: Optional[Dict[str, Any]],
        extensions: Optional[List[Extension]],
        allow_filter_test_or_use_query_overwrite: bool,
        embed_auditing_info: bool = False,
    ):  # pylint: disable=too-many-arguments
        super().__init__(
            loader=loader,  # nosec
            extensions=extensions,
            autoescape=select_autoescape(
                enabled_extensions=("htm", "html", "xml", "json"), default_for_string=False, default=False
            ),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
            lstrip_blocks=lstrip_blocks,
            trim_blocks=trim_blocks,
            optimized=False,  # cannot allow optimization because it prevents the use of stateful filters.
            auto_reload=False,
            cache_size=400,
        )
        if additional_globals is not None:
            for global_name, global_value in additional_globals.items():
                if global_name in self.RESERVED_GLOBAL_NAMESPACES or global_name in self.RESERVED_GLOBAL_NAMES:
                    raise RuntimeError(f'Additional global "{global_name}" uses a reserved global name')
                self.globals[global_name] = global_value

        self._allow_replacements = allow_filter_test_or_use_query_overwrite
        self._embed_auditing_info = embed_auditing_info

        for global_namespace in self.RESERVED_GLOBAL_NAMESPACES:
            self.globals[global_namespace] = LanguageTemplateNamespace()

        self.globals["now_utc"] = datetime.datetime(datetime.MINYEAR, 1, 1)

        self._additional_filters = additional_filters
        self._additional_tests = additional_tests
        self._target_language: Optional[Language] = None

    def set_language_context(self, lctx: LanguageContext) -> None:
        """
        Set the language context and update the environment with the supported languages.
        This also takes the target language from the context and updates the environment with the target language's
        globals, options, filters, and tests.
        """
        self._target_language = lctx.get_target_language()

        self._update_language_support(lctx)

        supported_languages = lctx.get_supported_languages().values()  # type: Optional[ValuesView[Language]]

        nunavut_namespace = self.nunavut_global
        setattr(nunavut_namespace, "embed_auditing_info", self._embed_auditing_info)
        setattr(nunavut_namespace, "platform_version", self._create_platform_version(self._embed_auditing_info))

        if self._target_language is None:
            omit_serialization = False
            support_version = (0, 0, 0)
            support_namespace = []
            target_language_name = "(no target language)"
            experimental = False
        else:
            omit_serialization = self._target_language.get_config_value_as_bool("omit_serialization_support", False)
            support_version = self._target_language.get_support_module()[1]
            support_namespace = self._target_language.support_namespace
            target_language_name = self._target_language.name
            experimental = not self._target_language.stable_support

        setattr(
            nunavut_namespace,
            "support",
            {
                "omit": omit_serialization,  # deprecated. Use "options.omit_serialization_support".
                "namespace": ".".join(support_namespace),
                "version": (support_version[0], support_version[1], support_version[2]),
            },
        )

        setattr(
            nunavut_namespace,
            "target_language",
            {"name": target_language_name, "experimental": experimental},
        )

        if "version" not in nunavut_namespace:
            # pylint: disable=import-outside-toplevel
            from nunavut import __version__ as nunavut_version

            setattr(nunavut_namespace, "version", nunavut_version)

        self.add_conventional_methods_to_environment(self)

        if self._additional_filters is not None:
            self._add_each_to_environment(
                self._additional_filters.items(), self.filters, supported_languages=supported_languages
            )
        if self._additional_tests is not None:
            self._add_each_to_environment(
                self._additional_tests.items(), self.tests, supported_languages=supported_languages
            )

    def add_conventional_methods_to_environment(self, obj: Any) -> None:
        """
        Adds methods using specific naming conventions to the Jinja environment. For example, methods named `filter_*`
        are added to the Jinja environment as filters.

        This method iterates over the methods of the given object and adds them to the Jinja environment.
        Only methods that are supported by the specified languages are added.

        :param Any obj: The object to add the methods from.

        """
        for name, method in inspect.getmembers(obj, inspect.isroutine):
            try:
                self._add_conventional_method_to_environment(method, name, supported_languages=self.supported_languages)
            except TypeError:
                pass

    @property
    def supported_languages(self) -> ValuesView[Language]:
        """
        The supported languages in the environment.

        :return: A view of the supported languages.
        :rtype: ValuesView[Language]
        """
        ln_globals = self.globals["ln"]  # type: LanguageTemplateNamespace
        return ln_globals.values()

    @property
    def nunavut_global(self) -> LanguageTemplateNamespace:
        """
        The `nunavut` global namespace.

        :return: The `nunavut` global namespace.
        :rtype: LanguageTemplateNamespace
        """
        return cast(LanguageTemplateNamespace, self.globals["nunavut"])

    @property
    def target_language_uses_queries(self) -> LanguageTemplateNamespace:
        """
        All `uses_queries` for the target language.

        :return: The uses queries for the target language.
        :rtype: LanguageTemplateNamespace
        """
        return cast(LanguageTemplateNamespace, self.globals["uses_queries"])

    @property
    def language_options(self) -> LanguageTemplateNamespace:
        """
        The language options.

        :return: The language options.
        :rtype: LanguageTemplateNamespace
        """
        return cast(LanguageTemplateNamespace, self.globals["options"])

    @property
    def language_support(self) -> LanguageTemplateNamespace:
        """
        The language support.

        :return: The language support.
        :rtype: LanguageTemplateNamespace
        """
        return cast(LanguageTemplateNamespace, self.globals["ln"])

    @property
    def target_language(self) -> Language:
        """
        The target language.

        :return: The target language.
        :rtype: Language
        """
        if self._target_language is None:
            raise RuntimeError("No target language has been set.")
        return self._target_language

    @property
    def now_utc(self) -> datetime.datetime:
        """
        Get or set the current UTC time.

        :return: The current UTC time.
        :rtype: datetime.datetime
        """
        return cast(datetime.datetime, self.globals["now_utc"])

    @now_utc.setter
    def now_utc(self, utc_time: datetime.datetime) -> None:
        self.globals["now_utc"] = utc_time

    @property
    def embed_auditing_info(self) -> bool:
        """
        Whether to embed auditing information in generated files.

        :return: Whether to embed auditing information.
        :rtype: bool
        """
        return self._embed_auditing_info

    def add_test(self, test_name: str, test_callable: Callable) -> None:
        """
        Add a test to the environment.

        :param str test_name: The name of the test.
        :param Callable test_callable: The test.
        :return: None
        """
        self._add_to_environment(test_name, test_callable, self.tests)

    # +----------------------------------------------------------------------------------------------------------------+
    # | Private
    # +----------------------------------------------------------------------------------------------------------------+
    def _resolve_collection(
        self,
        conventional_method_prefix: Optional[str],
        method_name: str,
        collection_maybe: Optional[Union[LanguageTemplateNamespace, Dict[str, Any]]],
    ) -> Union[LanguageTemplateNamespace, Dict[str, Any]]:
        """
        Resolve the collection to add the item to. If collection_maybe is not None then it is returned otherwise the
        collection is resolved based on the method name.
        """
        if collection_maybe is not None:
            return collection_maybe

        if LanguageEnvironment.is_test_name(conventional_method_prefix):
            return cast(Dict[str, Any], self.tests)
        if LanguageEnvironment.is_filter_name(conventional_method_prefix):
            return cast(Dict[str, Any], self.filters)
        if LanguageEnvironment.is_uses_query_name(conventional_method_prefix):
            uses_queries = self.globals["uses_queries"]
            return cast(LanguageTemplateNamespace, uses_queries)
        raise TypeError(
            f"Tried to add an item {method_name} to the template environment but we don't know what the item is."
        )

    def _add_to_environment(
        self,
        item_name: str,
        item: Any,
        collection: Union[LanguageTemplateNamespace, Dict[str, Any]],
    ) -> None:
        if item_name in collection:
            if not self._allow_replacements:
                raise RuntimeError(f"{item_name} was already defined.")
            if item_name in JINJA2_FILTERS:
                logger.info("Replacing Jinja built-in %s", item_name)
            else:
                logger.info('Replacing "%s" which was already defined for this environment.', item_name)
        else:
            logger.debug("Adding %s to environment", item_name)
        if isinstance(collection, LanguageTemplateNamespace):
            setattr(collection, item_name, item)
        else:
            collection[item_name] = item

    def _add_conventional_method_to_environment(
        self,
        method: Callable[..., bool],
        method_name: str,
        collection_maybe: Optional[Union[LanguageTemplateNamespace, Dict[str, Any]]] = None,
        supported_languages: Optional[ValuesView[Language]] = None,
        method_language: Optional[Language] = None,
        is_target: bool = False,
    ) -> None:
        """
        Add a method using specific naming conventions to the Jinja environment. For example, methods named `filter_*`
        are added to the Jinja environment as filters.

        :param Callable[..., bool] method: The named method.
        :param str method_name: The name of the callable to use in a template.
        :param Optional[Union[LanguageTemplateNamespace, Dict[str, Any]]] collection_maybe:
            The collection to add the method to. If None then the collection is resolved based on the method name.
        :param Optional[ValuesView[Language]] supported_languages: The supported languages.
        :param Optional[Language] method_language: The language of the method.
        :param bool is_target: Whether the method is for the target language.

            .. invisible-code-block: python

                from nunavut.jinja import CodeGenEnvironmentBuilder
                from nunavut.jinja.jinja2 import DictLoader
                from nunavut._templates import template_language_test
                from unittest.mock import MagicMock

                lctx = MagicMock(spec=LanguageContext)
                poop_lang = MagicMock(spec=Language)
                poop_lang.name = 'poop'
                poop_lang.get_templates_package_name = MagicMock(return_value='nunavut.lang.poop')
                lctx.get_target_language = poop_lang
                lctx.get_supported_languages = MagicMock(return_value = {'poop': poop_lang})

                @template_language_test('nunavut.lang.poop')
                def test_test(language):
                    return True

                e = (
                    CodeGenEnvironmentBuilder(DictLoader({'test': 'hello world'}))
                    .add_tests(foo=test_test)
                    .create(lctx)
                )
                assert test_test == e.tests['foo'].func
                assert e.tests['foo']()
        """

        result = LanguageEnvironment.handle_conventional_methods(method, method_name, supported_languages)
        collection = self._resolve_collection(result[0], method_name, collection_maybe)

        if method_language is not None:
            self._add_to_environment(f"ln.{method_language.name}.{result[1]}", result[2], collection)
        else:
            self._add_to_environment(result[1], result[2], collection)
        if is_target:
            self._add_to_environment(result[1], result[2], collection)

    def _add_each_to_environment(
        self,
        items: AbstractSet[Tuple[str, Callable]],
        collection: Optional[
            Union[
                LanguageTemplateNamespace,
                Dict[str, Any],
            ]
        ] = None,
        supported_languages: Optional[ValuesView[Language]] = None,
        language: Optional[Language] = None,
        is_target: bool = False,
    ) -> None:
        for method_name, method in items:
            self._add_conventional_method_to_environment(
                method, method_name, collection, supported_languages, language, is_target
            )

    @classmethod
    def _create_platform_version(cls, embed_auditing_info: bool) -> Dict[str, Any]:

        platform_version = {}  # type: Dict[str, Any]

        platform_version["python_version"] = platform.python_version()
        if embed_auditing_info:
            platform_version["python_implementation"] = platform.python_implementation()
            platform_version["python_release_level"] = sys.version_info[3]
            platform_version["python_build"] = platform.python_build()
            platform_version["python_compiler"] = platform.python_compiler()
            platform_version["python_revision"] = platform.python_revision()

            try:
                # pylint: disable=protected-access
                platform_version["python_xoptions"] = sys._xoptions
            except AttributeError:  # pragma: no cover
                platform_version["python_xoptions"] = {}

            platform_version["runtime_platform"] = platform.platform()

        return platform_version

    def _add_support_from_language_module_to_environment(
        self,
        lctx: LanguageContext,
        language: Language,
        ln_module: "types.ModuleType",
        is_target: bool = False,
    ) -> None:
        supported_languages = lctx.get_supported_languages()
        ln_env = LanguageEnvironment.find_all_conventional_methods_in_language_module(
            language, supported_languages.values(), ln_module
        )
        self._add_each_to_environment(
            ln_env.filters.items(), self.filters, supported_languages.values(), language=language, is_target=is_target
        )
        self._add_each_to_environment(
            ln_env.tests.items(), self.tests, supported_languages.values(), language=language, is_target=is_target
        )
        if is_target:
            self._target_language = language
            self._add_each_to_environment(
                ln_env.uses_queries.items(),
                self.globals["uses_queries"],
                supported_languages.values(),
                language=language,
                is_target=is_target,
            )

    def _update_language_support(self, lctx: LanguageContext) -> None:
        supported_languages = lctx.get_supported_languages()
        target_language = lctx.get_target_language()
        ln_globals = self.globals["ln"]
        self.globals.update(target_language.get_globals())
        globals_options_ns = self.globals["options"]
        globals_options_ns.update(target_language.get_options())
        for supported_language in supported_languages.values():
            if supported_language.name not in ln_globals:
                setattr(
                    ln_globals, supported_language.name, LanguageTemplateNamespace(options=LanguageTemplateNamespace())
                )
            ln_globals_ns = getattr(ln_globals, supported_language.name)
            ln_globals_ns.update(supported_language.get_globals())
            ln_globals_options_ns = getattr(ln_globals_ns, "options")
            ln_globals_options_ns.update(supported_language.get_options())

        # then load everything into the environment from this list.
        # note that we don't unload anything here so this method is not idempotent
        for supported_language in supported_languages.values():
            try:
                self._add_support_from_language_module_to_environment(
                    lctx,
                    supported_language,
                    LanguageClassLoader.load_language_module(supported_language.name),
                    (supported_language == target_language),
                )
            except ModuleNotFoundError:
                pass



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Language-specific support in nunavut.

This module contains the Language object and supporting types.
"""

import abc
import functools
import importlib
import logging
import pathlib
import types
import typing

import pydsdl

from nunavut._dependencies import Dependencies, DependencyBuilder
from nunavut._utilities import ResourceType, YesNoDefault, empty_list_support_files, iter_package_resources

from ._config import LanguageConfig, VersionReader

logger = logging.getLogger(__name__)


# +---------------------------------------------------------------------------+
# | ABSTRACT LANGUAGE
# +---------------------------------------------------------------------------+


class Language(metaclass=abc.ABCMeta):
    """
    Facilities for generating source code for a specific language. Concrete Language classes must be implemented
    by the language support package below lang and should be instantiated using
    :class:`nunavut.lang.LanguageClassLoader`.

    :param str module_name:                  The name of the :mod:`nunavut.lang` module that contains the concrete
                                             language type and its resources.
    :param LanguageConfig config:            All configuration as defined by the properties.json schema.
    :param kwargs:                           Opaque arguments passed through to the target
                                             :class:`nunavut.lang.Language` object. See all "WKLA" constants
                                             on this class for well-known keyword arguments.

        .. invisible-code-block: python

            from nunavut.lang._language import Language
            from nunavut.lang._language import _GenericLanguage
            from unittest.mock import MagicMock

            mock_config = MagicMock()

            try:
                my_lang = _GenericLanguage("foo", mock_config)
                # module must be within 'nunavut'
                assert False
            except ValueError:
                pass

            try:
                my_lang = _GenericLanguage("nunavut.foo", mock_config)
                # module must be within 'nunavut.lang'
                assert False
            except ValueError:
                pass

            try:
                my_lang = _GenericLanguage("not.nunavut.foo", mock_config)
                # module must be within 'nunavut.lang'
                assert False
            except ValueError:
                pass

            my_lang = _GenericLanguage("nunavut.lang.foo", mock_config)
            assert my_lang.name == 'foo'
    """

    # Well-Known Configuration Values (WKCV)
    # These are language configuration values the base Language class will look for.
    WKCV_DEFINITION_FILE_EXTENSION = "extension"
    WKCV_NAMESPACE_FILE_STEM = "namespace_file_stem"
    WKCV_SUPPORT_NAMESPACE = "support_namespace"
    WKCV_ENABLE_STROPPING = "enable_stropping"
    WKCV_HAS_STANDARD_NAMESPACE_FILES = "has_standard_namespace_files"
    WKCV_STABLE_SUPPORT = "stable_support"
    WKCV_NAMED_TYPES = "named_types"
    WKCV_NAMED_VALUES = "named_values"
    WKCV_LANGUAGE_OPTIONS = "options"
    WKCV_LANGUAGE_OPTION_DEFAULTS = "defaults"

    @classmethod
    def default_filter_id_for_target(cls, instance: typing.Any) -> str:
        """
        The default transformation of any object into a string.

        :param any instance:        Any object or data that either has a name property or can be converted to a string.
        :return: Either ``str(instance.name)`` if the instance has a name property or just ``str(instance)``
        """
        if hasattr(instance, "name"):
            return str(instance.name)
        else:
            return str(instance)

    # +-----------------------------------------------------------------------+
    # | LIFECYCLE AND DATA MODEL
    # +-----------------------------------------------------------------------+

    def __init__(
        self, language_module_name: str, config: LanguageConfig, **kwargs: typing.Any
    ):  # pylint: disable=unused-argument
        self._globals: typing.Optional[typing.Mapping[str, typing.Any]] = None
        self._section = language_module_name
        if not self._section.startswith(LanguageClassLoader.MODULE_PREFIX):
            raise ValueError(f"Unknown module name for language: {self._section}")
        self._language_name = LanguageClassLoader.to_language_name(self._section)
        self._config = config
        self._filters: typing.Dict[str, typing.Callable] = {}
        self._tests: typing.Dict[str, typing.Callable] = {}
        self._uses: typing.Dict[str, typing.Callable] = {}

        self._language_options = self._validate_language_options(
            config.get_config_value_as_dict(self._section, self.WKCV_LANGUAGE_OPTION_DEFAULTS, {}),
            config.get_config_value_as_dict(self._section, self.WKCV_LANGUAGE_OPTIONS, {}),
        )

    def __getattr__(self, name: str) -> typing.Any:
        """
        Any attribute access to a Language object will return the regular properties and
        any globals defined for the language. Because of this do not extend properties
        on this object in a way that will clash with the globals it defines (e.g. typename_
        or valuetoken_ should not be used as the start of attribute names).
        """
        try:
            return self.get_globals()[name]
        except KeyError as e:
            raise AttributeError(e) from e

    def __str__(self) -> str:
        return self._language_name

    def _validate_language_options(
        self, defaults: typing.Dict[str, typing.Any], options: typing.Dict[str, typing.Any]
    ) -> typing.Dict[str, typing.Any]:
        """
        Subclasses may override this method to validate language options. It will be invoked once
        by the base class constructor before setting the language options property.
        :param defaults: The a section of the language configuration that contains default values for options. The
                         format of this section is language-specific
        :param options: The options to validate.
        :return: The validated or modified options.
        :throws: ValueError if the options are invalid.

        .. invisible-code-block: python

            from nunavut.lang import LanguageContextBuilder
            from pytest import raises as assert_raises

            # test std
            language = LanguageContextBuilder()\
               .set_target_language("c")\
               .create()\
               .get_target_language()

            validated0 = language._validate_language_options(
                {},
                {
                    "std":"c11",
                }
            )

            assert validated0 is not None
            assert "std" in validated0

            validated1 = language._validate_language_options(
                {},
                {}
            )

            assert validated1 is not None

            validated2 = language._validate_language_options(
                {
                    "c11": {
                        "foo": "bar"
                    },
                },
                {
                    "std": "c11",
                    "foo": "baz"
                }
            )

            assert validated2 is not None
            assert "std" in validated2
            assert validated2["foo"] == "bar"

        """
        language_standard = options.get("std")

        if language_standard is not None and language_standard in defaults:
            options.update(defaults[language_standard])

        return options

    def _validate_globals(self, globals_map: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        """
        Subclasses may override this method to populate additional language-specific globals
        :param globals_map: The globals map to validate.
        :return: The validated or modified globals map.
        :throws: ValueError if the globals are invalid.
        """
        return globals_map

    # +-----------------------------------------------------------------------+
    # | PROPERTIES
    # +-----------------------------------------------------------------------+
    @property
    def extension(self) -> str:
        """
        The extension to use for files generated in this language.
        """
        return self._config.get_config_value(self._section, self.WKCV_DEFINITION_FILE_EXTENSION)

    @property
    def namespace_output_stem(self) -> typing.Optional[str]:
        """
        The name of a namespace file for this language.
        """
        return self._config.get_config_value(self._section, self.WKCV_NAMESPACE_FILE_STEM, None)

    @property
    def name(self) -> str:
        """
        The name of the language used by the :mod:`nunavut.lang` module.
        """
        return self._language_name

    @property
    def support_namespace(self) -> typing.List[str]:
        """
        The hierarchical namespace used by the support software. The property
        is a dot separated string when specified in configuration. This
        property returns that value split into namespace components with the
        first identifier being the first index in the array, etc.

        .. invisible-code-block: python

            from nunavut.lang import Language, LanguageContextBuilder

            lang_cpp = (
                LanguageContextBuilder(include_experimental_languages=True)
                    .set_target_language("cpp")
                    .set_target_language_configuration_override(Language.WKCV_SUPPORT_NAMESPACE, "foo.bar")
                    .create()
                    .get_target_language()
            )

            assert len(lang_cpp.support_namespace) == 2
            assert lang_cpp.support_namespace[0] == 'foo'
            assert lang_cpp.support_namespace[1] == 'bar'

        """
        namespace_str = self._config.get_config_value(self._section, self.WKCV_SUPPORT_NAMESPACE, default_value="")
        return namespace_str.split(".")

    @property
    def enable_stropping(self) -> bool:
        """
        Whether or not to strop identifiers for this language.
        """
        return self._config.get_config_value_as_bool(self._section, self.WKCV_ENABLE_STROPPING)

    @property
    def has_standard_namespace_files(self) -> bool:
        """
        Whether or not the language defines special namespace files as part of
        its core standard (e.g. python's __init__).
        """
        return self._config.get_config_value_as_bool(self._section, self.WKCV_HAS_STANDARD_NAMESPACE_FILES)

    @property
    def stable_support(self) -> bool:
        """
        Whether support for this language is designated 'stable', and not experimental.
        """
        return self._config.get_config_value_as_bool(self._section, self.WKCV_STABLE_SUPPORT)

    @property
    def named_types(self) -> typing.Mapping[str, str]:
        """
        Get a map of named types to the type name to emit for this language.
        """
        return self._config.get_config_value_as_dict(self._section, self.WKCV_NAMED_TYPES, default_value={})

    @property
    def named_values(self) -> typing.Mapping[str, str]:
        """
        Get a map of named values to the token to emit for this language.
        """
        return self._config.get_config_value_as_dict(self._section, self.WKCV_NAMED_VALUES, default_value={})

    # +-----------------------------------------------------------------------+
    # | METHODS
    # +-----------------------------------------------------------------------+

    def get_support_module(self) -> typing.Tuple[str, typing.Tuple[int, int, int], typing.Optional["types.ModuleType"]]:
        """
        Returns the module object for the language support files.
        :return: A tuple of module name, x.y.z module version, and the module object itself.

        .. invisible-code-block: python

            from nunavut.lang._language import Language, _GenericLanguage, LanguageConfig
            from unittest.mock import MagicMock

            mock_config = MagicMock()
            mock_module_name= LanguageClassLoader.to_language_module_name("cpp")

            my_lang = _GenericLanguage(mock_module_name, mock_config)
            my_lang._section = LanguageClassLoader.to_language_module_name("cpp")
            module_name, support_version, _ = my_lang.get_support_module()

            assert module_name == LanguageClassLoader.to_language_module_name("cpp") + ".support"
            assert support_version[0] == 1

        """
        module_name = f"{self._section}.support"

        try:
            module = importlib.import_module(module_name)
            version_tuple = VersionReader.read_version(module)
            return (module_name, version_tuple, module)
        except (ImportError, ValueError):
            # No serialization support for this language
            logger.info("No serialization support for selected target. Cannot retrieve module.")

            return (module_name, (0, 0, 0), None)

    @functools.lru_cache()
    def get_dependency_builder(self, for_type: pydsdl.Any) -> DependencyBuilder:
        """
        Get a dependency builder for the given type.
        """
        return DependencyBuilder(for_type)

    @abc.abstractmethod
    def get_includes(self, dep_types: Dependencies) -> typing.List[str]:
        """
        Get a list of include paths that are specific to this language and the options set for it.
        :param Dependencies dep_types: A description of the dependencies includes are needed for.
        :return: A list of include file paths. The list may be empty if no includes were needed.
        """

    def filter_id(self, instance: typing.Any, id_type: str = "any") -> str:  # pylint: disable=unused-argument
        """
        Produces a valid identifier in the language for a given object. The encoding may not be reversible.

        :param any instance:        Any object or data that either has a name property or can be converted
                                    to a string.
        :param str id_type:         A type of identifier. This is different for each language. For example, for C this
                                    value can be 'typedef', 'macro', 'function', or 'enum'.
                                    Use 'any' to apply stropping rules for all identifier types to the instance.
        :return: A token that is a valid identifier in the language, is not a reserved keyword, and is transformed
                in a deterministic manner based on the provided instance.
        """
        return self.default_filter_id_for_target(instance)

    def filter_short_reference_name(
        self, t: pydsdl.CompositeType, stropping: YesNoDefault = YesNoDefault.DEFAULT, id_type: str = "any"
    ) -> str:
        """
        Provides a string that is a shorted version of the full reference name omitting any namespace parts of the type.

        :param pydsdl.CompositeType t: The DSDL type to get the reference name for.
        :param YesNoDefault stropping: If DEFAULT then the stropping value configured for the target language is used
                                       else this overrides that value.
        :param str id_type:         A type of identifier. This is different for each language. For example, for C this
                                    value can be 'typedef', 'macro', 'function', or 'enum'.
                                    Use 'any' to apply stropping rules for all identifier types to the instance.
        """
        short_name = f"{t.short_name}_{t.version.major}_{t.version.minor}"
        if YesNoDefault.test_truth(stropping, self.enable_stropping):
            return self.filter_id(short_name, id_type)
        return short_name

    def get_config_value(self, key: str, default_value: typing.Optional[str] = None) -> str:
        """
        Get an optional language property from the language configuration.

        :param str key: The config value to retrieve.
        :param default_value: The value to return if the key was not in the configuration. If provided
                this method will not raise.
        :type default_value: typing.Optional[str]
        :return: Either the value from the config or the default_value if provided.
        :rtype: str
        :raises: KeyError if the section or the key in the section does not exist and a default_value was not provided.
        """
        return self._config.get_config_value(self._section, key, default_value)

    def get_config_value_as_bool(self, key: str, default_value: bool = False) -> bool:
        """
        Get an optional language property from the language configuration returning a boolean.

        :param str key: The config value to retrieve.
        :param bool default_value: The value to use if no value existed.
        :return: The config value as either True or False.
        :rtype: bool
        """
        return self._config.get_config_value_as_bool(self._section, key, default_value)

    def get_config_value_as_dict(
        self, key: str, default_value: typing.Optional[typing.Dict] = None
    ) -> typing.Dict[str, typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        .. invisible-code-block: python

            from nunavut.lang._language import LanguageConfig, LanguageClassLoader, Language, _GenericLanguage

            config = LanguageConfig()
            config.add_section('nunavut.lang.c')
            config.set('nunavut.lang.c', 'foo', {'one': 1})

            lang_c = _GenericLanguage(LanguageClassLoader.to_language_module_name('c'), config)

            assert lang_c.get_config_value_as_dict('foo')['one'] == 1

            assert lang_c.get_config_value_as_dict('bar', {'one': 2})['one'] == 2

        :param str key:           The config value to retrieve.
        :param default_value:     The value to return if the key was not in the configuration. If provided this method
                                  will not raise a KeyError nor a TypeError.
        :type default_value: typing.Optional[typing.Mapping[str, typing.Any]]
        :return:                  Either the value from the config or the default_value if provided.
        :rtype: typing.Mapping[str, typing.Any]
        :raises: KeyError if the key does not exist and a default_value was not provided.
        :raises: TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        return self._config.get_config_value_as_dict(self._section, key, default_value)

    def get_config_value_as_list(
        self, key: str, default_value: typing.Optional[typing.List] = None
    ) -> typing.List[typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        :param str key:           The config value to retrieve.
        :param default_value:     The value to return if the key was not in the configuration. If provided this method
                                  will not raise a KeyError nor a TypeError.
        :type default_value:      typing.Optional[typing.List[typing.Any]]
        :return:                  Either the value from the config or the default_value if provided.
        :rtype:                   typing.List[typing.Any]
        :raises:                  KeyError if the key does not exist and a default_value was not provided.
        :raises:                  TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        return self._config.get_config_value_as_list(self._section, key, default_value)

    def get_support_files(
        self, resource_type: int = ResourceType.ANY.value
    ) -> typing.Generator[pathlib.Path, None, None]:
        """
        Iterates over supporting files embedded within the Nunavut distribution.

        :param resource_type: The type of support resources to enumerate.

        .. invisible-code-block: python

            from nunavut.lang._language import Language, _GenericLanguage, LanguageClassLoader
            from unittest.mock import MagicMock

            mock_config = MagicMock()
            mock_module_name_ = LanguageClassLoader.to_language_module_name("foo")

            my_lang = _GenericLanguage(mock_module_name_, mock_config)
            my_lang._section =  LanguageClassLoader.to_language_module_name("not_a_language_really_not_a_language")
            for support_file in my_lang.get_support_files():
                # if the module doesn't exist it shouldn't have any support files.
                assert False

        """
        _, _, module = self.get_support_module()

        if module is not None:
            # All language support modules must provide a list_support_files method
            # to allow the copy generator access to the packaged support files.
            list_support_files: typing.Callable[[int], typing.Generator[pathlib.Path, None, None]] = getattr(
                module, "list_support_files"
            )
            return list_support_files(resource_type)
        else:
            return empty_list_support_files()

    def get_option(
        self, option_key: str, default_value: typing.Union[typing.Mapping[str, typing.Any], str, bool, None] = None
    ) -> typing.Union[typing.Mapping[str, typing.Any], str, bool, None]:
        """
        Get a language option for this language.

        .. invisible-code-block: python

            from nunavut.lang import Language, LanguageContextBuilder

            options = {'target_endianness': 'little'}

            lang_cpp = (
                LanguageContextBuilder(include_experimental_languages=True)
                    .set_target_language("cpp")
                    .set_target_language_configuration_override(Language.WKCV_LANGUAGE_OPTIONS, options)
                    .create()
                    .get_target_language()
            )

        .. code-block:: python

            # Values can come from defaults...
            assert lang_cpp.get_option('target_endianness') == 'little'

            # ... or can come from a sane default.
            assert lang_cpp.get_option('foobar', 'sane_default') == 'sane_default'

        :return: Either the value provided to the :class:`nunavut.lang.Language` instance, the value from
            properties.json or the :code:`default_value`.

        """
        try:
            return self._language_options[option_key]  # type: ignore
        except KeyError:
            return default_value

    def get_templates_package_name(self) -> str:
        """
        The name of the nunavut python package containing filters, types, and configuration
        for this language.
        """
        return self._section

    def get_globals(self) -> typing.Mapping[str, typing.Any]:
        """
        Get all values for this language that should be available in a global context.

        :return: A mapping of global names to global values.
        """

        # If this method is called before init (see __getattr__) then we raise an AttributeError to avoid
        # infinite recursion
        _ = object.__getattribute__(self, "_globals")

        if self._globals is None:
            globals_map: typing.Dict[str, typing.Any] = {}

            for key, value in self.named_types.items():
                globals_map[f"typename_{key}"] = value
            for key, value in self.named_values.items():
                globals_map[f"valuetoken_{key}"] = value

            self._globals = self._validate_globals(globals_map)
        return self._globals

    def get_options(self) -> typing.Mapping[str, typing.Any]:
        """
        Get all language options for this Language.

        :return: A mapping of option names to option values.
        """
        return self._language_options


# +---------------------------------------------------------------------------+
# | UNDEFINED LANGUAGE TYPE (concrete)
# +---------------------------------------------------------------------------+


class _GenericLanguage(Language):
    """
    Language type used when the language support within Nunavut does not define a language-specific
    subclass.

    Do not use this. Use :class:`nunavut.lang.LanguageClassLoader` which will create the proper object type
    based on inspection of the Nunavut internals.
    """

    def get_includes(self, dep_types: Dependencies) -> typing.List[str]:
        return []


# +---------------------------------------------------------------------------+
# | LANGUAGE CLASSLOADER
# +---------------------------------------------------------------------------+


class LanguageClassLoader:
    """
    Class loader to resolve concrete :class:`nunavut.lang.Language` types.

    :param additional_config_files: A list of paths to additional configuration files to load as configuration.
        These will override any values found in the :file:`nunavut.lang.properties.json` file and files
        appearing later in this list will override value found in earlier entries.

        .. invisible-code-block: python

            from nunavut.lang import LanguageClassLoader, LanguageConfig

            subject = LanguageClassLoader()

            c_reserved_identifiers = subject.config.get_config_value(
                LanguageClassLoader.to_language_module_name("c"), "reserved_identifiers")
            assert len(c_reserved_identifiers) > 0
    """

    MODULE_NAME = "nunavut.lang"
    MODULE_PREFIX = MODULE_NAME + "."
    MODULE_FORMAT = MODULE_PREFIX + "{}"

    @classmethod
    def to_language_name(cls, unknown_string: str) -> str:
        """
        Helper method to take a string that is either a language name or a language module name
        and always return a language name.

        .. invisible-code-block: python
            from nunavut.lang import LanguageClassLoader

        .. code-block: python

            assert "c" == LanguageClassLoader.to_language_name("c")
            assert "c" == LanguageClassLoader.to_language_name("nunavut.lang.c")

        """
        return (
            unknown_string[len(cls.MODULE_PREFIX) :] if unknown_string.startswith(cls.MODULE_PREFIX) else unknown_string
        )

    @classmethod
    def to_language_module_name(cls, unknown_string: str) -> str:
        """
        Helper method to take a string that is either a language name or a language module name
        and always return a language module name.

        .. invisible-code-block: python
            from nunavut.lang import LanguageClassLoader

        .. code-block: python

            assert "nunavut.lang.c" == LanguageClassLoader.to_language_module_name("c")
            assert "nunavut.lang.c" == LanguageClassLoader.to_language_module_name("nunavut.lang.c")

        """
        return (
            cls.MODULE_FORMAT.format(unknown_string)
            if not unknown_string.startswith(cls.MODULE_PREFIX)
            else unknown_string
        )

    @classmethod
    def _load_config(cls) -> LanguageConfig:
        parser = LanguageConfig()
        for resource in iter_package_resources(cls.MODULE_NAME, ".json"):
            ini_string = resource.read_text()
            parser.update_from_json_string(ini_string)
        return parser

    def __init__(self) -> None:
        self._config: typing.Optional[LanguageConfig] = None

    @classmethod
    def load_language_module(cls, language_name: str) -> types.ModuleType:
        """
        Load a language module by name.
        """
        module_name = cls.to_language_module_name(language_name)
        return importlib.import_module(module_name)

    @property
    def config(self) -> LanguageConfig:
        """
        Meta-data about all languages merged from the Nunavut internal defaults and any additional
        configuration files provided to this class's constructor.
        """
        if self._config is None:
            self._config = self._load_config()
        return self._config

    @functools.lru_cache()
    def load_language_class(self, language_name: str) -> typing.Tuple[types.ModuleType, typing.Type[Language]]:
        """
        :param str language_name:                The name of the language used by the :mod:`nunavut.lang` module.

        :return: A concrete :class:`nunavut.lang.Language`.

        .. invisible-code-block: python

            from nunavut.lang import LanguageClassLoader, Language

            lang_c_mod, lang_c_cls = LanguageClassLoader().load_language_class('c')
            assert lang_c_mod.__name__ == 'nunavut.lang.c'
            assert issubclass(lang_c_cls, Language)

        """
        ln_module = self.load_language_module(language_name)

        try:
            language_type = typing.cast(typing.Type["Language"], getattr(ln_module, "Language"))
        except AttributeError:
            logging.debug(
                "Unable to find a Language object in nunavut.lang.%s. Using a Generic language object", language_name
            )
            language_type = _GenericLanguage

        if language_type == Language:
            # the language module just imported the base class so let's go ahead and use _GenericLanguage
            language_type = _GenericLanguage

        return (ln_module, language_type)

    def new_language(self, language_name: str, **kwargs: typing.Any) -> Language:
        """
        Instantiate a new language class. This is a helper that reuses the internally held
        module and :class:`LanguageConfig` to invoke the required :class:`nunavut.lang.Language` constructor.

        :param str language_name:                The name of the language used by the :mod:`nunavut.lang` module.
        :param kwargs:                           Opaque arguments passed to the :class:`nunavut.lang.Language`
                                                 constructor.

        :return: A new :class:`nunavut.lang.Language` instance.

        .. invisible-code-block: python

            from nunavut.lang import LanguageClassLoader

        .. code-block:: python

            # Using this helper method:

            lang_c = LanguageClassLoader().new_language('c')

            assert lang_c.name == 'c'


            # Without the helper:

            loader = LanguageClassLoader()
            lang_c_module, lang_c_class = loader.load_language_class('c')
            lang_c = lang_c_class(lang_c_module.__name__, loader.config)

            assert lang_c.name == 'c'

        .. invisible-code-block: python

            # let's go ahead and load the rest of our known, internally supported languages just to raise
            # test failures right here at the wellspring.

            lang_py = LanguageClassLoader().new_language('py')
            assert lang_py.name == 'py'

            lang_js = LanguageClassLoader().new_language('js')
            assert lang_js.name == 'js'

        """
        ln_module, ln_class = self.load_language_class(language_name)
        return ln_class(ln_module.__name__, config=self.config, **kwargs)



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Command line script.
"""

import sys

from .cli.runners import main as cli_main

sys.exit(cli_main())



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Empty python package to ensure the support generator doesn't explode.
"""

import pathlib
import typing

from nunavut._utilities import ResourceType, empty_list_support_files

__version__ = "1.0.0"
"""Version of the js support headers."""


def list_support_files(resource_type: int = ResourceType.ANY.value) -> typing.Generator[pathlib.Path, None, None]:
    """
    Get a list of javascript support files embedded in this package.
    :param resource_type: A type of support file to list.
    """
    # pylint: disable=unused-argument
    return empty_list_support_files()



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Language-specific support in nunavut.

This package contains modules that provide specific support for generating
source for various languages using templates.
"""

import functools
import logging
import pathlib
import typing

from ._config import LanguageConfig
from ._language import Language, LanguageClassLoader

logger = logging.getLogger(__name__)


class UnsupportedLanguageError(ValueError):
    """
    Error type raised if an unsupported language type is used.
    """


class LanguageContextBuilder:
    """
    Used to instantiate new :class:`LanguageContext` objects.

    The simplest invocation will always work by using the :data:`LanguageContextBuilder.DEFAULT_TARGET_LANGUAGE`
    constant:

    .. code-block:: python

        from nunavut.lang import LanguageContextBuilder

        default_language_context = LanguageContextBuilder().create()

        assert LanguageContextBuilder.DEFAULT_TARGET_LANGUAGE == default_language_context.get_target_language().name

    Typically a target language is specified at minimum. Also see constants on :class:`nunavut.lang.Language` for
    well-known options that the builder can override:

    .. code-block:: python

        from nunavut.lang import LanguageContextBuilder

        customized_language_context = (
            LanguageContextBuilder()
                .set_target_language("c")
                .set_target_language_configuration_override(Language.WKCV_DEFINITION_FILE_EXTENSION, ".h")
                .create()
        )

        assert customized_language_context.get_target_language().extension == ".h"

    :param include_experimental_languages: If set then languages that are not fully supported will be allowed otherwise
        any experimental languages will be missing and errors will be raised as if the language specified was unknown.

    """

    DEFAULT_TARGET_LANGUAGE = "c"  #: The target language used for new contexts if none is specified.

    def __init__(self, include_experimental_languages: bool = False):
        self._target_language_name: typing.Optional[str] = None
        self._target_language_config: typing.Dict[str, typing.Any] = {}
        self._ln_loader = LanguageClassLoader()
        self._include_experimental_languages = include_experimental_languages

    def get_supported_language_names(self) -> typing.Iterable[str]:
        """
        Get a list of target languages supported by Nunavut.

        :return: An iterable of strings which are languages with special
            support within Nunavut templates.
        """
        return [LanguageClassLoader.to_language_name(s) for s in self._ln_loader.config.sections()]

    @property
    def config(self) -> LanguageConfig:
        """
        The configuration object that will be used to create the language context.
        """
        return self._ln_loader.config

    # +-----------------------------------------------------------------------+
    # | BUILDER SYNTAX
    # +-----------------------------------------------------------------------+

    def set_target_language_configuration_override(self, key: str, value: typing.Any) -> "LanguageContextBuilder":
        """
        Stores a key and value to override in the configuration for a language target when a LanguageContext is crated.
        These overrides are always set under the language section of the target language.

        .. invisible-code-block: python

            from nunavut.lang import LanguageContextBuilder, Language, LanguageClassLoader

        .. code-block:: python

            builder = LanguageContextBuilder().set_target_language("c")

            default_c_file_extension = builder.config.get_config_value(
                                            LanguageClassLoader.to_language_module_name("c"),
                                            Language.WKCV_DEFINITION_FILE_EXTENSION)

            assert default_c_file_extension == ".h"

        We can now try to override the file extension for a future "C" target language object:

        .. code-block:: python

            builder.set_target_language_configuration_override(Language.WKCV_DEFINITION_FILE_EXTENSION, ".foo")

        ...but that value will not be overridden until you create the target language:

        .. code-block:: python

            default_c_file_extension = builder.config.get_config_value(
                                            LanguageClassLoader.to_language_module_name("c"),
                                            Language.WKCV_DEFINITION_FILE_EXTENSION)

            assert default_c_file_extension == ".h"

            _ = builder.create()

            overridden_c_file_extension = builder.config.get_config_value(
                                            LanguageClassLoader.to_language_module_name("c"),
                                            Language.WKCV_DEFINITION_FILE_EXTENSION)

            assert overridden_c_file_extension == ".foo"

        Note that the config is scoped by the builder but is then inherited by the language objects created by the
        builder:

        .. code-block:: python

            one = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .set_target_language_configuration_override("foo", 1)
                )
            two = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .set_target_language_configuration_override("foo", 2)
                )

            # Here we see that the second override of "foo" does not affect the first because they
            # are in different builders.

            assert (
                    one.create().get_target_language().get_config_value("foo")
                    !=
                    two.create().get_target_language().get_config_value("foo")
                )

        """
        if value is not None:
            self._target_language_config[key] = value
        return self

    def add_target_language_option_override(self, key: str, value: typing.Any) -> "LanguageContextBuilder":
        """
        Adds a key and value to override in the language options section of the configuration for a language target
        when a LanguageContext is crated. These overrides are always set under the language section of the target
        language.

        .. invisible-code-block: python

            from nunavut.lang import LanguageContextBuilder, Language, LanguageClassLoader

        .. code-block:: python

            builder = LanguageContextBuilder().set_target_language("c")
            c_section_name = LanguageClassLoader.to_language_module_name("c")

            default_c_language_options = builder.config.get_config_value_as_dict(
                                            c_section_name,
                                            Language.WKCV_LANGUAGE_OPTIONS)

            assert default_c_language_options["std"] == "c11"

        We can now try to override the default standard for a future "C" target language object:

        .. code-block:: python

            builder.add_target_language_option_override("std", "c17")

        ...but that value will not be overridden until you create the target language:

        .. code-block:: python


            default_c_language_options = builder.config.get_config_value_as_dict(
                                            c_section_name,
                                            Language.WKCV_LANGUAGE_OPTIONS)

            assert default_c_language_options["std"] == "c11"

            _ = builder.create()

            overridden_c_language_options = builder.config.get_config_value_as_dict(
                                            c_section_name,
                                            Language.WKCV_LANGUAGE_OPTIONS)

            assert default_c_language_options["std"] == "c17"

        Note that the config is scoped by the builder but is then inherited by the language objects created by the
        builder in the same way as the configuration overrides.

        .. invisible-code-block: python

            from pytest import raises

            builder = LanguageContextBuilder().set_target_language("c")
            c_section_name = LanguageClassLoader.to_language_module_name("c")
            builder.set_target_language_configuration_override(
                Language.WKCV_LANGUAGE_OPTIONS,
                "wrong type"
            )
            with raises(ValueError):
                builder.add_target_language_option_override("std", "c17")
            # This will raise a ValueError because the WKCV_LANGUAGE_OPTIONS is not a dict.

        """
        if value is not None:
            options = self._target_language_config.get(Language.WKCV_LANGUAGE_OPTIONS)
            if options is None:
                options = {}
                self._target_language_config[Language.WKCV_LANGUAGE_OPTIONS] = options
            if not isinstance(options, dict):
                raise ValueError(
                    f"Cannot set target language option override for {key} because the value is not a dict."
                )
            options[key] = value
        return self

    def set_target_language_extension(
        self, target_language_extension: typing.Optional[str]
    ) -> "LanguageContextBuilder":
        """
        Helper method for setting the target language file extension (since this is a common override).

        Calling this method is the same as doing:

        .. invisible-code-block: python

            from nunavut.lang import LanguageContextBuilder, Language, LanguageClassLoader

        .. code-block:: python

            LanguageContextBuilder().set_target_language_configuration_override(
                Language.WKCV_DEFINITION_FILE_EXTENSION,
                ".h")

        """
        return self.set_target_language_configuration_override(
            Language.WKCV_DEFINITION_FILE_EXTENSION, target_language_extension
        )

    def set_target_language(self, target_language: typing.Optional[str]) -> "LanguageContextBuilder":
        """
        Set the language name to target. This can be either the name of the language, as defined by Nunavut, or
        it can be the language package name.

        .. invisible-code-block: python

            from nunavut.lang import LanguageContextBuilder, LanguageClassLoader

        .. code-block:: python

            assert LanguageContextBuilder().set_target_language("c").create().get_target_language().name == "c"

            assert (
                LanguageContextBuilder()
                    .set_target_language(LanguageClassLoader.to_language_module_name("c"))
                    .create()
                    .get_target_language().name == "c"
            )

        Also note that, if the language name is None, the default name will be assigned internally:

        .. code-block:: python

            target_language = LanguageContextBuilder().set_target_language(None).create().get_target_language()

            assert target_language.name == LanguageContextBuilder.DEFAULT_TARGET_LANGUAGE

        """
        if target_language is None:
            self._target_language_name = self.DEFAULT_TARGET_LANGUAGE
        else:
            self._target_language_name = LanguageClassLoader.to_language_name(target_language)
        return self

    def set_additional_config_files(
        self, additional_config_files: typing.List[pathlib.Path]
    ) -> "LanguageContextBuilder":
        """
        Deprecated. Use :func:`add_config_files` instead.
        """
        logger.warning("set_additional_config_files is deprecated. Use add_config_files instead.")
        return self.add_config_files(*additional_config_files)

    def add_config_files(self, *additional_config_files: pathlib.Path) -> "LanguageContextBuilder":
        """
        A list of paths to additional json files to load as configuration.
        These will override any values found in the :file:`nunavut.lang.properties.json` file and files
        appearing later in this list will override value found in earlier entries.

        .. invisible-code-block: python

            import pathlib
            import json
            import textwrap
            from nunavut.lang import LanguageContextBuilder, Language, LanguageClassLoader

            overrides_file = gen_paths_for_module.out_dir / pathlib.Path("overrides1.json")

            overrides_data = {
                LanguageClassLoader.to_language_module_name("c"): {Language.WKCV_DEFINITION_FILE_EXTENSION: ".foo"}
            }

            with open(overrides_file, "w", encoding="utf-8") as overrides_handle:
                json.dump(overrides_data, overrides_handle)

        .. code-block:: python

            target_language_w_overrides = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .add_config_files(overrides_file)
                    .create()
                    .get_target_language()
            )

            target_language_no_overrides = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .create()
                    .get_target_language()
            )

            assert target_language_w_overrides.extension == ".foo"
            assert target_language_no_overrides.extension == ".h"

        Overrides are applied as unions. For example, given this override data:

        .. code-block:: python

            overrides_data = '''
                {
                    "nunavut.lang.c" :
                        {
                            "extension": ".foo",
                            "non-standard": "bar"
                        }
                }
            '''

        ...the standard "extension" property will be overridden and the "non-standard" property will be added.

        .. invisible-code-block: python

            second_overrides_file = gen_paths_for_module.out_dir / pathlib.Path("overrides2.json")
            with open(second_overrides_file, "w", encoding="utf-8") as overrides_handle:
                overrides_handle.write(textwrap.dedent(overrides_data))

        .. code-block:: python

            target_language_w_overrides = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .add_config_files(second_overrides_file)
                    .create()
                    .get_target_language()
            )

            assert ".foo" == target_language_w_overrides.extension
            assert "bar" == target_language_w_overrides.get_config_value("non-standard")

        .. invisible-code-block: python

            from nunavut import DefaultValue

            # verification of issue #329 fix
            with_default = {"enable_serialization_asserts" : DefaultValue(False) }
            without_default = {"enable_serialization_asserts" : False }

            # verification of issue #329 fix
            overrides_data = '''
                nunavut.lang.c:
                    options:
                        enable_serialization_asserts: true
            '''

            issue_329_overrides = gen_paths_for_module.out_dir / pathlib.Path("overrides329.yaml")
            with open(issue_329_overrides, "w", encoding="utf-8") as overrides_handle:
                overrides_handle.write(textwrap.dedent(overrides_data))

            target_language_329_no_file_overrides = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .set_target_language_configuration_override(Language.WKCV_LANGUAGE_OPTIONS, with_default)
                    .create()
                    .get_target_language()
            )

            target_language_329_file_override = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .add_config_files(issue_329_overrides)
                    .set_target_language_configuration_override(Language.WKCV_LANGUAGE_OPTIONS, with_default)
                    .create()
                    .get_target_language()
            )

            target_language_329_file_override_overridden = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .add_config_files(issue_329_overrides)
                    .set_target_language_configuration_override(Language.WKCV_LANGUAGE_OPTIONS, without_default)
                    .create()
                    .get_target_language()
            )

            # default from command line
            assert not target_language_329_no_file_overrides.get_option("enable_serialization_asserts")

            # default from command line overridden by file
            assert target_language_329_file_override.get_option("enable_serialization_asserts")

            # command-line overrides file
            assert not target_language_329_file_override_overridden.get_option("enable_serialization_asserts")
        """
        for additional_path in additional_config_files:
            if additional_path.suffix not in (".json", ".yaml", ".yml"):
                raise ValueError(f"Unsupported file type {additional_path.suffix} for configuration file.")
            with additional_path.open("r", encoding="utf-8") as additional_file:
                if additional_path.suffix == ".json":
                    self.config.update_from_json_file(additional_file)
                else:
                    self.config.update_from_yaml_file(additional_file)

        return self

    def create(self) -> "LanguageContext":
        """
        Applies all pending configuration overrides to the internal :class:`LanguageConfig` object and instantiates
        a :class:`LanguageContext` object.
        """
        # First find the target language to use...
        target_language_name = self._resolve_target_language(self._target_language_name)

        # Now update the configuration for the target language with everything we stored in this
        # builder instance...
        self.config.update_section(
            LanguageClassLoader.to_language_module_name(target_language_name), self._target_language_config
        )

        # Create the target language instance...
        target_language = self._new_language_w_experimental_handling(target_language_name)

        # and finally, build the LanguageContext.
        return LanguageContext(
            self._ln_loader.config, target_language, functools.partial(self._new_language_map, target_language)
        )

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    def _new_language_w_experimental_handling(self, language_name: str) -> Language:
        try:
            language = self._ln_loader.new_language(language_name)
        except ImportError as e:
            logger.debug("Import Error %s when trying to load language %s", e, language_name)
            raise UnsupportedLanguageError(f"language {language_name} is not a supported language") from e
        if not (language.stable_support or self._include_experimental_languages):
            raise UnsupportedLanguageError(
                f"{language_name} support is only experimental, but experimental language support is not enabled"
            )
        return language

    def _new_language_map(self, target_language: Language) -> typing.Dict[str, Language]:
        """
        Build a map of all supported languages.
        :param target_language: The target language is included in the returned map but must be build
            by another method.
        """
        languages: typing.Dict[str, Language] = {target_language.name: target_language}
        for language_name in set(self.get_supported_language_names()) - set((target_language.name,)):
            try:
                languages[language_name] = self._new_language_w_experimental_handling(language_name)
            except UnsupportedLanguageError:
                pass
        return languages

    def _resolve_target_language(self, explicit_value: typing.Optional[str]) -> str:
        if explicit_value is not None:
            return explicit_value

        inferred_target_language_name: typing.Optional[str] = None

        target_extension = self._target_language_config.get(Language.WKCV_DEFINITION_FILE_EXTENSION, None)
        if target_extension is not None:
            for language_config_section_name, language_config_section in self.config.sections().items():
                if language_config_section.get("extension", None) == target_extension:
                    inferred_target_language_name = LanguageClassLoader.to_language_name(language_config_section_name)
                    break

        if inferred_target_language_name is None:
            inferred_target_language_name = self.DEFAULT_TARGET_LANGUAGE
            logger.info(
                "No target language specified and none could be inferred. Using default language, %s",
                self.DEFAULT_TARGET_LANGUAGE,
            )
        else:
            logging.info(
                'Inferring target language %s based on extension "%s".',
                inferred_target_language_name,
                target_extension,
            )
        return inferred_target_language_name


class LanguageContext:
    """
    Context object containing the current target language and all supported :class:`nunavut.lang.Language` objects.

    :param language_configuration: The configuration for all languages as defined by the properties.json schema.
    :param target_language: The target language.
    :param supported_language_builder: factory closure that will create :class:`nunavut.lang.Language` objects for
                                       all supported languages when :func:`LanguageContext.get_target_languages`
                                       is first called.
    """

    def __init__(
        self,
        language_configuration: LanguageConfig,
        target_language: Language,
        supported_language_builder: typing.Callable[[], typing.Dict[str, Language]],
    ):
        self._config = language_configuration
        self._target_language = target_language
        self._all_supported_languages: typing.Optional[typing.Dict[str, Language]] = None
        self._all_supported_languages_builder = supported_language_builder

    def get_language(self, key_or_module_name: str) -> Language:
        """
        Get a :class:`nunavut.lang.Language` object for a given language identifier.

        :param str key_or_module_name: Either one of the Nunavut mnemonics for a supported language or
            the ``__name__`` of one of the ``nunavut.lang.[language]`` python modules.
        :return: A :class:`nunavut.lang.Language` object cached by this context.
        :rtype: nunavut.lang.Language
        """
        if key_or_module_name is None or len(key_or_module_name) == 0:
            raise ValueError("key argument is required.")
        key = LanguageClassLoader.to_language_name(key_or_module_name)
        return self.get_supported_languages()[key]

    def get_target_language(self) -> Language:
        """
        Returns the target language for code generation.
        """
        return self._target_language

    def filter_id_for_target(self, instance: typing.Any, id_type: str = "any") -> str:
        """
        A filter that will transform a given string or pydsdl identifier into a valid identifier in the target language.

        :param any instance:        Any object or data that either has a name property or can be converted to a string.
        :param str id_type:         A type of identifier. This is different for each language.
                                    Use 'any' to apply stropping rules for all identifier types to the instance.
        :return: A token that is a valid identifier in the target language, is not a reserved keyword, and is
                 transformed in a deterministic manner based on the provided instance.
        """
        return self._target_language.filter_id(instance, id_type)

    def get_supported_languages(self) -> typing.Dict[str, Language]:
        """
        Returns a collection of available language support objects.

         .. invisible-code-block: python

            from nunavut.lang import LanguageContextBuilder

            lctx = LanguageContextBuilder().create()

            default_language = None

            for language_name, language in lctx.get_supported_languages().items():
                if language_name == LanguageContextBuilder.DEFAULT_TARGET_LANGUAGE:
                    default_language = language
                    break

            assert default_language is not None
            assert default_language.name == LanguageContextBuilder.DEFAULT_TARGET_LANGUAGE
            assert len(lctx.get_supported_languages()) > 1

        """
        if self._all_supported_languages is None:
            self._all_supported_languages = self._all_supported_languages_builder()
        return self._all_supported_languages

    @property
    def config(self) -> LanguageConfig:
        """
        Returns the :class:`nunavut.lang.LanguageConfig` object that contains the configuration for all
        supported languages. This is the same object that is used to instantiate the :class:`nunavut.lang.Language`
        """
        return self._config
